from pytest import approx

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.registration.demo.demo_translation_3d import _register_translation_3d


def test_register_translation_3d_numpy():
    with NumpyBackend():
        register_translation_3d()


def test_register_translation_3d_cupy():
    try:
        with CupyBackend():
            register_translation_3d()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def register_translation_3d(length_xy=128):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    image, shifted, unshifted, model = _register_translation_3d(length_xy=length_xy, display=False)
    shifts = Backend.to_numpy(model.shift_vector)
    assert shifts[0] == approx(-1, abs=0.2)
    assert shifts[1] == approx(-5, abs=0.2)
    assert shifts[2] == approx(13, abs=0.2)



from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.utils.speed_test import perform_speed_test


def test_backend_context():
    perform_speed_test()



import click
from arbol.arbol import aprint, asection

from dexp.cli.parsing import _parse_channels, _parse_slicing
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.view import dataset_view


@click.command()
@click.argument('input_paths', nargs=-1)
@click.option('--channels', '-c', default=None, help='list of channels, all channels when ommited.')
@click.option('--slicing', '-s', default=None, help='dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z).')
@click.option('--aspect', '-a', type=float, default=4, help='sets aspect ratio e.g. 4', show_default=True)
@click.option('--clim', '-cl', type=str, default='0,512', help='Sets the contrast limits, i.e. -cl 0,1000 sets the contrast limits to [0,1000]', show_default=True)
@click.option('--colormap', '-cm', type=str, default='viridis', help='sets colormap, e.g. viridis, gray, magma, plasma, inferno ', show_default=True)
@click.option('--windowsize', '-ws', type=int, default=1536, help='Sets the napari window size. i.e. -ws 400 sets the window to 400x400', show_default=True)
@click.option('--projectionsonly', '-po', is_flag=True, help='To view only the projections, if present.', show_default=True)
@click.option('--volumeonly', '-vo', is_flag=True, help='To view only the volumetric data.', show_default=True)
def view(input_paths,
         channels,
         slicing,
         aspect,
         clim,
         colormap,
         windowsize,
         projectionsonly,
         volumeonly
         ):
    """ Views dataset using napari (napari.org)
    """

    slicing = _parse_slicing(slicing)

    name = input_paths[0] + '...' if len(input_paths) > 1 else ''

    contrast_limits = list(float(v.strip()) for v in clim.split(','))

    if len(input_paths) == 1 and 'http' in input_paths[0]:

        with asection(f"Viewing dataset at: {input_paths}, channels: {channels}, slicing: {slicing}, aspect:{aspect} "):
            dataset_view(input_paths,
                         name,
                         aspect,
                         channels,
                         contrast_limits,
                         colormap,
                         slicing,
                         windowsize)
            aprint("Done!")


    else:
        input_dataset, input_paths = glob_datasets(input_paths)
        channels = _parse_channels(input_dataset, channels)

        with asection(f"Viewing dataset at: {input_paths}, channels: {channels}, slicing: {slicing}, aspect:{aspect} "):
            dataset_view(input_dataset=input_dataset,
                         name=name,
                         aspect=aspect,
                         channels=channels,
                         contrast_limits=contrast_limits,
                         colormap=colormap,
                         slicing=slicing,
                         windowsize=windowsize,
                         projections_only=projectionsonly,
                         volume_only=volumeonly)
            input_dataset.close()
            aprint("Done!")



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.filters.kernels.gaussian import gaussian_kernel_nd
from dexp.processing.filters.kernels.wiener import wiener_kernel


def demo_wiener_numpy():
    with NumpyBackend():
        _demo_wiener()


def demo_wiener_cupy():
    try:
        with CupyBackend():
            _demo_wiener()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_wiener():
    xp = Backend.get_xp_module()

    psf = gaussian_kernel_nd(size=31,
                             ndim=2,
                             dtype=numpy.float32)

    psf_f = xp.log1p(xp.absolute(xp.fft.fftshift(xp.fft.fftn(psf))))

    w = wiener_kernel(kernel=psf)

    w_f = xp.log1p(xp.absolute(xp.fft.fftshift(xp.fft.fftn(w))))

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(psf), name='psf')
        viewer.add_image(_c(psf_f), name='psf_f', colormap='viridis')
        viewer.add_image(_c(w), name='b')
        viewer.add_image(_c(w_f), name='b_f', colormap='viridis')
        viewer.grid.enabled = True
        viewer.grid.view = (2, 2)


demo_wiener_cupy()
demo_wiener_numpy()



from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.utils.fit_shape import fit_to_shape


def test_fit_shape_numpy():
    with NumpyBackend():
        _test_fit_shape()


def test_fit_shape_cupy():
    try:
        with CupyBackend():
            _test_fit_shape()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_fit_shape():
    xp = Backend.get_xp_module()

    array_1 = xp.random.uniform(0, 1, size=(31, 10, 17))
    array_2 = xp.random.uniform(0, 1, size=(32, 9, 18))

    array_2_fit = fit_to_shape(array_2.copy(), shape=array_1.shape)

    assert array_2_fit is not array_2
    assert array_2_fit.shape == array_1.shape

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     viewer = Viewer()
    #     viewer.add_image(array_1, name='array_1')
    #     viewer.add_image(array_2, name='array_2')
    #     viewer.add_image(array_2_fit, name='array_2_fit')



from typing import Union, Tuple, Sequence

from arbol import aprint

from dexp.processing.backends.backend import Backend
from dexp.processing.color.blend import blend_color_images
from dexp.processing.color.border import add_border


def insert_color_image(image,
                       inset_image,
                       scale: Union[float, Tuple[float, ...]] = 1,
                       translation: Union[str, Sequence[Tuple[Union[int, float], ...]]] = None,
                       border_width: int = 0,
                       border_color: Tuple[float, float, float, float] = None,
                       border_over_image: bool = False,
                       mode: str = 'max',
                       alpha: float = 1,
                       background_color: Tuple[float, float, float, float] = (0, 0, 0, 0),
                       rgba_value_max: float = 255):
    """
    Inserts an inset image into a base image.
    After scaling the inset image must be smaller than the base image.

    Parameters
    ----------
    image: Base image.
    inset_image: Inset image to place in base image.
    scale: scale factor for inset image -- scaling happens before translation.
    translation: positions of the insets in pixels in natural order: (x, y). Can also be a string: 'bottom_left', 'bottom_right', 'top_left', 'top_right'.
    width: Width of border added to insets.
    color: Border color.
    over_image: If True the border is not added but overlayed over the image, the image does not change size.
    mode: blending mode.
    alpha: inset transparency.
    background_color: background color as tuple of normalised floats:  (R,G,B,A). Default is transparent black.
    rgba_value_max: Max value for the pixel/voxel values.

    Returns
    -------
    Image with inset inserted.

    """

    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    # Move to backend:
    image = Backend.to_backend(image)
    inset_image = Backend.to_backend(inset_image)

    # Normalise scale:
    if type(scale) == int or type(scale) == float:
        scale = (scale,) * (inset_image.ndim - 1) + (1,)

    scale = tuple(scale)
    if len(scale) == inset_image.ndim - 1:
        scale = scale + (1,)

    # Scale inset image:
    if any(s != 1 for s in scale):
        aprint(f"Scaling image by: {scale}")
        inset_image = sp.ndimage.zoom(inset_image, zoom=scale)

    # Add border:
    if border_width != 0:
        aprint(f"Adding border of width: {border_width}, and color: {border_color}")
        inset_image = add_border(inset_image,
                                 width=border_width,
                                 color=border_color,
                                 over_image=border_over_image,
                                 rgba_value_max=rgba_value_max,
                                 )

    # Check that if after scaling, the inset image is smaller than the base image:
    if any(u > v for u, v in zip(inset_image.shape, image.shape)):
        raise ValueError(f"Inset image {inset_image.shape} too large compared to base image {image.shape}.")

    # Check that both images have the same number of dimensions:
    if inset_image.ndim != image.ndim:
        raise ValueError(f"Inset image number of dimensions: {inset_image.ndim} does not match base image number of dimensions: {image.ndim}.")

    # Pad inset image:
    pad_width = tuple((v - u for u, v in zip(inset_image.shape, image.shape)))
    pad_width = pad_width[0:-1] + (0,)
    pad_width = tuple((0, p) for p in pad_width)
    padded_inset_image = xp.pad(inset_image, pad_width=pad_width)

    # Roll inset image to place it at the correct position:
    axis = tuple(range(image.ndim))
    shift = tuple((0,) * image.ndim)
    if translation is None:
        # Nothing to do...
        pass
    elif type(translation) == str:
        shift = list(shift)
        if 'top' in translation:
            shift[0] += 0
        elif 'bottom' in translation:
            shift[0] += image.shape[0] - inset_image.shape[0]
        if 'left' in translation:
            shift[1] += 0
        elif 'right' in translation:
            shift[1] += image.shape[1] - inset_image.shape[1]
        shift = tuple(shift)
    elif type(translation) == tuple:
        shift = translation + (0,)
    else:
        raise ValueError(f"Unsupported translation: {translation}")
    # cast to int:
    shift = tuple(int(round(u)) for u in shift)

    # translate:
    padded_inset_image = xp.roll(padded_inset_image, axis=axis, shift=shift)

    # Blend images together:
    result = blend_color_images(images=(image, padded_inset_image),
                                alphas=(1, alpha),
                                modes=('max', mode),
                                background_color=background_color,
                                rgba_value_max=rgba_value_max)

    return result



from arbol import aprint

from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.video.demo.demo_blend import demo_blend


def test_blend_numpy():
    with NumpyBackend():
        demo_blend(display=False)


def test_blend_cupy():
    try:
        with CupyBackend():
            demo_blend(display=False)

    except ModuleNotFoundError:
        aprint("Cupy module not found! demo ignored")



# !/usr/bin/env python

"""
the actual renderer class to max project 3d data

the modelView and projection matrices are compatible with OpenGL

usage:

rend = VolumeRenderer((400,400))

Nx,Ny,Nz = 200,150,50
d = linspace(0,10000,Nx*Ny*Nz).reshape([Nz,Ny,Nx])

rend.set_data(d)
rend.set_units([1.,1.,.1])
rend.set_projection(projMatPerspective(60,1.,1,10))
rend.set_modelView(dot(mat4_translate(0,0,-7),mat4_scale(.7,.7,.7)))

out = rend.color()



author: Martin Weigert
email: mweigert@mpi-cbg.de
"""

import logging

from six.moves import zip

from dexp.volumerender.transform_matrices import mat4_perspective, mat4_identity, mat4_scale

logger = logging.getLogger(__name__)

import os
# this is due to some pyinstaller bug!
from scipy.linalg import inv
from time import time
import sys
from gputools import init_device, get_device, OCLProgram, OCLArray, OCLImage
import numpy as np


def absPath(myPath):
    """ Get absolute path to resource, works for dev and for PyInstaller """
    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        base_path = sys._MEIPASS
        return os.path.join(base_path, os.path.basename(myPath))
    except Exception:
        base_path = os.path.abspath(os.path.dirname(__file__))
        return os.path.join(base_path, myPath)


class VolumeRenderer:
    """ renders a data volume by ray casting/max projection

    usage:
               rend = VolumeRenderer((400,400))
               rend.set_data(d)
               rend.set_units(1.,1.,2.)
               rend.set_modelView(rotMatX(.7))
    """
    dtypes = [np.float32, np.uint16, np.uint8]
    interpolation_defines = {"linear": ["-D", "SAMPLER_FILTER=CLK_FILTER_LINEAR"],
                             "nearest": ["-D", "SAMPLER_FILTER=CLK_FILTER_NEAREST"]}

    def __init__(self, size=None, interpolation='linear'):
        """ e.g. size = (300,300)"""

        self.__DEFAULTMAXSTEPS__ = 200
        self.__QUALIFIER_CONSTANT_TO_GLOBAL__ = 0
        try:
            # simulate GPU fail...
            # raise Exception()

            # self.dev = OCLDevice(useGPU = True,
            #                      useDevice = spimagine.__OPENCLDEVICE__)

            # FIXME
            # device should not be reinitialized as then
            # every other queue becomes invalid
            # init_device(useGPU = True,
            #             useDevice = spimagine.config.__OPENCLDEVICE__)
            self.isGPU = True


        except Exception as e:
            print(e)
            print("could not find GPU OpenCL device -  trying CPU...")

            try:
                init_device(useGPU=False)
                self.isGPU = False
                self.dtypes = [np.float32]
            except Exception as e:
                print(e)
                print("could not find any OpenCL device ... sorry")

        self.memMax = .7 * get_device().get_info("MAX_MEM_ALLOC_SIZE")

        # self.memMax = 2.*get_device().get_info("MAX_MEM_ALLOC_SIZE")

        self.rebuild_program(interpolation=interpolation)

        self.invMBuf = OCLArray.empty(16, dtype=np.float32)

        self.invPBuf = OCLArray.empty(16, dtype=np.float32)

        self.projection = np.zeros((4, 4))
        self.modelView = np.zeros((4, 4))

        if size:
            self.resize(size)
        else:
            self.resize((200, 200))

        self.set_dtype()
        self.set_gamma()
        self.set_max_val()
        self.set_min_val()

        self.set_occ_strength(.1)
        self.set_occ_radius(21)
        self.set_occ_n_points(30)

        self.set_alpha_pow()
        self.set_box_boundaries()
        self.set_units()

        self.set_modelView()
        self.set_projection()

    def rebuild_program(self, interpolation="linear"):
        build_options_basic = ["-I", "%s" % absPath("kernels/"),
                               "-D", "maxSteps=%s" % self.__DEFAULTMAXSTEPS__,

                               ]

        if self.__QUALIFIER_CONSTANT_TO_GLOBAL__:
            build_options_basic += ["-D", "QUALIFIER_CONSTANT_TO_GLOBAL"]

        if interpolation in VolumeRenderer.interpolation_defines:
            build_options_basic += VolumeRenderer.interpolation_defines[interpolation]
        else:
            raise KeyError(
                "interpolation = '%s' not defined ,valid: %s" % (interpolation, list(VolumeRenderer.interpolation_defines.keys())))

        try:
            self.proc = OCLProgram(absPath("kernels/all_render_kernels.cl"),
                                   build_options=
                                   build_options_basic +
                                   ["-cl-finite-math-only",
                                    "-cl-fast-relaxed-math",
                                    "-cl-unsafe-math-optimizations",
                                    "-cl-mad-enable"])
        except Exception as e:
            logger.debug(str(e))
            self.proc = OCLProgram(absPath("kernels/all_render_kernels.cl"),
                                   build_options=
                                   build_options_basic)
        self.proc

    def set_dtype(self, dtype=None):
        if hasattr(self, "dtype") and dtype is self.dtype:
            return

        if dtype is None:
            dtype = self.dtypes[0]

        if dtype in self.dtypes:
            self.dtype = dtype
        else:
            raise NotImplementedError("data type should be either %s not %s" % (self.dtypes, dtype))

        self.reset_buffer()

    def resize(self, size):
        self.width, self.height = size
        self.reset_buffer()

    def reset_buffer(self):
        self.buf = OCLArray.empty((self.height, self.width), dtype=np.float32)
        self.buf_alpha = OCLArray.empty((self.height, self.width), dtype=np.float32)
        self.buf_depth = OCLArray.empty((self.height, self.width), dtype=np.float32)
        self.buf_normals = OCLArray.empty((self.height, self.width, 3), dtype=np.float32)
        self.buf_tmp = OCLArray.empty((self.height, self.width), dtype=np.float32)
        self.buf_tmp_vec = OCLArray.empty((self.height, self.width, 3), dtype=np.float32)

        self.buf_occlusion = OCLArray.empty((self.height, self.width), dtype=np.float32)

        self.output = np.zeros((self.height, self.width), dtype=np.float32)
        self.output_alpha = np.zeros((self.height, self.width), dtype=np.float32)
        self.output_depth = np.zeros((self.height, self.width), dtype=np.float32)

    def _get_downsampled_data_slices(self, data):
        """in case data is bigger then gpu texture memory, we should downsample it
        if so returns the slice of data to be rendered
        else returns None (no downsampling)
        """
        # Nstep = int(np.ceil(np.sqrt(1.*data.nbytes/self.memMax)))
        Nstep = int(np.ceil((1. * data.nbytes / self.memMax) ** (1. / 3)))

        slices = [slice(0, d, Nstep) for d in data.shape]
        if Nstep > 1:
            logger.info("downsample image by factor of  %s" % Nstep)
            return slices
        else:
            return None

    def set_max_val(self, maxVal=0.):
        self.maxVal = maxVal

    def set_min_val(self, minVal=0.):
        self.minVal = minVal

    def set_gamma(self, gamma=1.):
        self.gamma = gamma

    def set_occ_strength(self, occ=.2):
        self.occ_strength = occ

    def set_occ_radius(self, rad=21):
        self.occ_radius = rad

    def set_occ_n_points(self, n_points=31):
        self.occ_n_points = n_points

    def set_alpha_pow(self, alphaPow=0.):
        self.alphaPow = alphaPow

    def set_data(self, data, autoConvert=True, copyData=False):
        logger.debug("set_data")

        if not autoConvert and not data.dtype in self.dtypes:
            raise NotImplementedError("data type should be either %s not %s" % (self.dtypes, data.dtype))

        if data.dtype.type in self.dtypes:
            self.set_dtype(data.dtype.type)
            _data = data
        else:
            print("converting type from %s to %s" % (data.dtype.type, self.dtype))
            _data = data.astype(self.dtype, copy=False)

        self.dataSlices = self._get_downsampled_data_slices(_data)

        if self.dataSlices is not None:
            self.set_shape(_data[self.dataSlices].shape[::-1])
        else:
            self.set_shape(_data.shape[::-1])

        t = time()
        self.update_data(_data, copyData=copyData)
        logger.debug("update data: %s ms" % (1000. * (time() - t)))
        self.update_matrices()

    def set_shape(self, dataShape):
        if self.isGPU:
            self.dataImg = OCLImage.empty(dataShape[::-1], dtype=self.dtype)
        else:
            raise NotImplementedError("TODO")
            # self.dataImg = self.dev.createImage(dataShape,
            #     mem_flags = cl.mem_flags.READ_ONLY,
            #     channel_order = cl.channel_order.INTENSITY,
            #     channel_type = cl_datatype_dict[self.dtype])

            # if self.isGPU:
            #     self.dataImg = self.dev.createImage(dataShape,
            #         mem_flags = cl.mem_flags.READ_ONLY,
            #         channel_type = cl_datatype_dict[self.dtype])
            # else:
            #     self.dataImg = self.dev.createImage(dataShape,
            #         mem_flags = cl.mem_flags.READ_ONLY,
            #         channel_order = cl.channel_order.INTENSITY,
            #         channel_type = cl_datatype_dict[self.dtype])

    def update_data(self, data, copyData=False):
        # do we really want to copy here?

        if self.dataSlices is not None:
            self._data = data[self.dataSlices].copy()
        else:
            if copyData:
                self._data = data.copy()
            else:
                self._data = data

        if self._data.dtype != self.dtype:
            self._data = self._data.astype(self.dtype, copy=False)

        self.dataImg.write_array(self._data)

    def set_box_boundaries(self, boxBounds=[-1, 1, -1, 1, -1, 1]):
        self.boxBounds = np.array(boxBounds)

    def set_units(self, stackUnits=np.ones(3)):
        self.stackUnits = np.array(stackUnits)

    def set_projection(self, projection=mat4_perspective()):
        self.projection = projection
        self.update_matrices()

    def set_modelView(self, modelView=mat4_identity()):
        self.modelView = 1. * modelView
        self.update_matrices()

    def update_matrices(self):
        if hasattr(self, "dataImg"):
            mScale = self._stack_scale_mat()
            invM = inv(np.dot(self.modelView, mScale))
            self.invMBuf.write_array(invM.flatten().astype(np.float32))
            invP = inv(self.projection)
            self.invPBuf.write_array(invP.flatten().astype(np.float32))

    def _stack_scale_mat(self):
        # scaling the data according to size and units
        Nx, Ny, Nz = self.dataImg.shape
        dx, dy, dz = self.stackUnits

        # mScale =  scaleMat(1.,1.*dx*Nx/dy/Ny,1.*dx*Nx/dz/Nz)
        maxDim = max(d * N for d, N in zip([dx, dy, dz], [Nx, Ny, Nz]))
        return mat4_scale(1. * dx * Nx / maxDim, 1. * dy * Ny / maxDim, 1. * dz * Nz / maxDim)

    def _render_max_project(self, dtype=np.float32, numParts=1, currentPart=0):
        if dtype in [np.uint16, np.uint8]:
            method = "max_project_short"
        elif dtype == np.float32:

            method = "max_project_float"
        else:
            raise NotImplementedError("wrong dtype: %s", dtype)

        self.proc.run_kernel(method,
                             (self.width, self.height),
                             None,
                             self.buf.data, self.buf_alpha.data,
                             self.buf_depth.data,
                             np.int32(self.width), np.int32(self.height),
                             np.float32(self.boxBounds[0]),
                             np.float32(self.boxBounds[1]),
                             np.float32(self.boxBounds[2]),
                             np.float32(self.boxBounds[3]),
                             np.float32(self.boxBounds[4]),
                             np.float32(self.boxBounds[5]),
                             np.float32(self.minVal),
                             np.float32(self.maxVal),
                             np.float32(self.gamma),
                             np.float32(self.alphaPow),
                             np.int32(numParts),
                             np.int32(currentPart),
                             self.invPBuf.data,
                             self.invMBuf.data,
                             self.dataImg)

        self.output = self.buf.get()
        self.output_alpha = self.buf_alpha.get()
        self.output_depth = self.buf_depth.get()

    def _convolve_scalar(self, buf, radius=11):

        self.proc.run_kernel("conv_x",
                             (self.width, self.height), None,
                             buf.data,
                             self.buf_tmp.data,
                             np.int32(radius))
        self.proc.run_kernel("conv_y",
                             (self.width, self.height), None,
                             self.buf_tmp.data,
                             buf.data,
                             np.int32(radius))

    def _convolve_vec(self, buf, radius=11):
        self.proc.run_kernel("conv_vec_x",
                             (self.width, self.height), None,
                             buf.data,
                             self.buf_tmp_vec.data,
                             np.int32(radius))

        self.proc.run_kernel("conv_vec_y",
                             (self.width, self.height), None,
                             self.buf_tmp_vec.data,
                             buf.data,
                             np.int32(radius))

    def _render_isosurface2(self):
        self.proc.run_kernel("iso_surface",
                             (self.width, self.height),
                             None,
                             self.buf.data, self.buf_alpha.data,
                             self.buf_depth.data, self.buf_normals.data,
                             np.int32(self.width), np.int32(self.height),
                             np.float32(self.boxBounds[0]),
                             np.float32(self.boxBounds[1]),
                             np.float32(self.boxBounds[2]),
                             np.float32(self.boxBounds[3]),
                             np.float32(self.boxBounds[4]),
                             np.float32(self.boxBounds[5]),
                             np.float32(self.maxVal / 2),
                             np.float32(self.gamma),
                             self.invPBuf.data,
                             self.invMBuf.data,
                             self.dataImg,
                             np.int32(self.dtype in [np.uint16, np.uint8])
                             )

        self._convolve_vec(self.buf_normals, 5)

        self.output = self.buf.get()
        self.output_alpha = self.buf_alpha.get()
        self.output_depth = self.buf_depth.get()
        self.output_normals = self.buf_normals.get()

    def _render_isosurface(self):
        """
        with ambient occlusion
        """

        self.proc.run_kernel("iso_surface",
                             (self.width, self.height),
                             None,
                             self.buf.data, self.buf_alpha.data,
                             self.buf_depth.data, self.buf_normals.data,
                             np.int32(self.width), np.int32(self.height),
                             np.float32(self.boxBounds[0]),
                             np.float32(self.boxBounds[1]),
                             np.float32(self.boxBounds[2]),
                             np.float32(self.boxBounds[3]),
                             np.float32(self.boxBounds[4]),
                             np.float32(self.boxBounds[5]),
                             np.float32(self.maxVal / 2),
                             np.float32(self.gamma),
                             self.invPBuf.data,
                             self.invMBuf.data,
                             self.dataImg,
                             np.int32(self.dtype in [np.uint16, np.uint8])
                             )
        self._convolve_vec(self.buf_normals, 7)
        #
        self.proc.run_kernel("occlusion",
                             (self.width, self.height),
                             None,
                             self.buf_occlusion.data,
                             np.int32(self.width), np.int32(self.height),
                             np.int32(self.occ_radius),
                             np.int32(self.occ_n_points),
                             self.buf_depth.data,
                             self.buf_normals.data,
                             )

        self._convolve_scalar(self.buf_occlusion, 5)

        self.proc.run_kernel("shading",
                             (self.width, self.height),
                             None,
                             self.buf.data, self.buf_alpha.data,
                             np.int32(self.width), np.int32(self.height),
                             self.invPBuf.data,
                             self.invMBuf.data,
                             np.float32(self.occ_strength),
                             self.buf_normals.data,
                             self.buf_depth.data,
                             self.buf_occlusion.data,

                             )

        # self._convolve_scalar(self.buf,13)
        # self._convolve_vec(self.buf_normals,101)

        self.output = self.buf.get()
        self.output_alpha = self.buf_alpha.get()
        self.output_depth = self.buf_depth.get()
        self.output_normals = self.buf_normals.get()
        self.output_occlusion = self.buf_occlusion.get()

    def render(self, data=None, stackUnits=None,
               minVal=None, maxVal=None, gamma=None,
               modelView=None, projection=None,
               boxBounds=None, return_alpha=False, method="max_project",
               numParts=1, currentPart=0):

        if data is not None:
            self.set_data(data)

        if maxVal is not None:
            self.set_max_val(maxVal)

        if minVal is not None:
            self.set_min_val(minVal)

        if gamma is not None:
            self.set_gamma(gamma)

        if stackUnits is not None:
            self.set_units(stackUnits)

        if modelView is not None:
            self.set_modelView(modelView)

        if projection is not None:
            self.set_projection(projection)

        if not hasattr(self, 'dataImg'):
            print("no data provided, set_data(data) before")
            return

        if modelView is None and not hasattr(self, 'modelView'):
            print("no modelView provided and set_modelView() not called before!")
            return

        if method == "max_project":
            self._render_max_project(self.dtype, numParts, currentPart)

        if method == "iso_surface":
            self._render_isosurface()



import math
from typing import Tuple, Union, Any

from dexp.processing.backends.backend import Backend
from dexp.processing.utils.nd_slice import nd_split_slices


def scatter_gather_i2v(function,
                       images: Union[Any, Tuple[Any]],
                       chunks: Union[int, Tuple[int, ...]],
                       margins: Union[int, Tuple[int, ...]] = None,
                       to_numpy: bool = True,
                       internal_dtype=None):
    """
    Image-2-vector scatter-gather.
    Given a n-ary function that takes n images and returns a tuple of vectors per image, we split the input arrays into chunks,
    apply the function to each chunk computing using a given backend, and reassembling the vectors into a tuple of arrays of vectors,
    with one vector per chunk.

    Parameters
    ----------
    function : n-ary function. Must accept one or more arrays -- of same shape -- and return a _tuple_ of arrays as result.
    images : sequence of images (can be any backend, numpy )
    chunks : chunks to cut input image into, can be a single integer or a tuple of integers.
    margins : margins to add to each chunk, can be a single integer or a tuple of integers.
    to_numpy : should the result be a numpy array? Very usefull when the compute backend cannot hold the whole input and output images in memory.
    internal_dtype : internal dtype for computation

    Returns
    -------
    Result a tuple of arrays, each array having the dimension of the input images + an extra 'vector' dimensions.
    The shape of the arrays matches the chunking of the input images. Each vector is the result of applying the
    function to the input images. If to_numpy==True then the returned arrays are numpy array.

    """

    xp = Backend.get_xp_module()

    if type(images) is not tuple:
        images = (images,)

    for image in images:
        if image.shape != images[0].shape:
            raise ValueError("All images must have the same shape!")

    first_image = images[0]
    ndim = first_image.ndim
    shape = first_image.shape
    dtype = first_image.dtype

    if internal_dtype is None:
        internal_dtype = dtype

    if type(chunks) == int:
        chunks = (chunks,) * ndim

    if margins is None:
        margins = (0,) * ndim
    elif type(margins) == int:
        margins = (margins,) * ndim

    # We compute the slices objects to cut the input image into batches:
    chunk_slices = list(nd_split_slices(shape, chunks=chunks, margins=margins))
    chunk_slices_no_margins = list(nd_split_slices(shape, chunks=chunks))

    # We compute the shape at the chunk level:
    chunk_shape = tuple(math.ceil(s / c) for s, c in zip(shape, chunks))

    # Zipping together slices with and without margins:
    slices = zip(chunk_slices, chunk_slices_no_margins)

    # Number of tiles:
    number_of_tiles = len(chunk_slices)

    if number_of_tiles == 1:
        # If there is only one tile, let's not be complicated about it:
        results = function(*images)
        if to_numpy:
            results = tuple(Backend.to_numpy(result, dtype=dtype) for result in results)
        else:
            results = tuple(Backend.to_backend(result, dtype=dtype) for result in results)
        results_stacked_reshaped = tuple(xp.reshape(result, newshape=(1,) * ndim + result.shape) for result in results)
    else:
        results_lists = None
        for chunk_slice, chunk_slice_no_margins in slices:
            image_chunks = tuple(image[chunk_slice] for image in images)
            image_chunks = tuple(Backend.to_backend(image_chunk, dtype=internal_dtype) for image_chunk in image_chunks)
            results = function(*image_chunks)
            if to_numpy:
                results = tuple(Backend.to_numpy(result, dtype=image.dtype) for result in results)
            else:
                results = tuple(Backend.to_backend(result, dtype=image.dtype) for result in results)

            if results_lists is None:
                results_lists = tuple([] for _ in results)

            for result, results_list in zip(results, results_lists):
                results_list.append(result)

        rxps = tuple(Backend.get_xp_module(results_list[0]) for results_list in results_lists)
        results_stacked = tuple(rxp.stack(results_list) for rxp, results_list in zip(rxps, results_lists))
        results_stacked_reshaped = tuple(rxp.reshape(stack, newshape=chunk_shape + results[0].shape) for rxp, stack, results in zip(rxps, results_stacked, results_lists))

    return results_stacked_reshaped



from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend


def fuse_dct_nd(image_a,
                image_b,
                cutoff: float = 0,
                clip: bool = True,
                internal_dtype=None):
    """
    Fuses two images in DCT domain by picking coefficients with maximal magnitude.

    Parameters
    ----------
    image_a : First image to fuse
    image_b : Second image to fuse
    cutoff : frequency cutoff
    clip : Clip fused image to the min and max voxel values of the original images
    -- i.e. the fused image must remain within the bounds of the original images.
    internal_dtype : dtype for internal computation

    Returns
    -------
    Fused image.

    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if not image_a.shape == image_b.shape:
        raise ValueError("Arrays must have the same shape")

    if not image_a.dtype == image_b.dtype:
        raise ValueError("Arrays must have the same dtype")

    if not type(Backend.current()) is NumpyBackend:
        raise NotImplementedError("DCT not yet implemented in Cupy")

    if internal_dtype is None:
        internal_dtype = image_a.dtype

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = xp.float32

    original_dtype = image_a.dtype

    image_a = Backend.to_backend(image_a, dtype=internal_dtype)
    image_b = Backend.to_backend(image_b, dtype=internal_dtype)

    min_a, max_a = xp.min(image_a), xp.max(image_a)
    min_b, max_b = xp.min(image_b), xp.max(image_b)
    min_value = min(min_a, min_b)
    max_value = min(max_a, max_b)

    tranform = lambda x: sp.fft.dctn(x, norm='ortho')
    itranform = lambda x: sp.fft.idctn(x, norm='ortho')

    image_a_dct = tranform(image_a)
    image_b_dct = tranform(image_b)

    image_a_dct_abs = xp.absolute(image_a_dct)
    image_b_dct_abs = xp.absolute(image_b_dct)
    image_a_is_max = image_a_dct_abs > image_b_dct_abs

    image_fused_dct = image_a_is_max * image_a_dct + ~image_a_is_max * image_b_dct

    if 0 < cutoff <= 1:
        c = cutoff
        cutoffs = tuple(int(s * c) for s in image_a.shape)
        cutoffs_slice = tuple(slice(0, c) for c in cutoffs)
        image_fused_dct[cutoffs_slice] = ~image_a_is_max[cutoffs_slice] * image_a_dct[cutoffs_slice] \
                                         + image_a_is_max[cutoffs_slice] * image_b_dct[cutoffs_slice]

    image_fused = itranform(image_fused_dct)

    if clip:
        image_fused = xp.clip(image_fused, min_value, max_value, out=image_fused)

    # Adjust type:
    image_fused = image_fused.astype(original_dtype, copy=False)

    return image_fused



from typing import Tuple

import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.utils.element_wise_affine import element_wise_affine


def normalise_functions(image,
                        low: float = 0.,
                        high: float = 1.,
                        minmax: Tuple[float, float] = None,
                        quantile: float = 0,
                        clip: bool = False,
                        do_normalise: bool = True,
                        in_place: bool = True,
                        dtype=None):
    """ Returns a pair of functions: the first normalises the given image to the range [low, high], the second denormalises back to the original range (and dtype).
    Usefull when determining the normalisation parameters and doing the actual normalisation must be decoupled, for example when splitting an image into chunks and sensing computation to the GPU.


    Parameters
    ----------
    image : image to normalise
    low, high : normalisation range
    minmax : min and max values of the image if already known.
    quantile : if quantile>0 then quantile normalisation is used to find the min and max values. Value must be within [0,1]
    clip : clip after normalisation/denormalisation
    do_normalise : If False, the returned functions are pass-through identity functions -- usefull to turn on and off normalisation while still using the functions themselves.
    in_place : In-place computation is allowed and inputs may be modified
    dtype : dtype to normalise to.

    Returns
    -------

    """

    xp = Backend.get_xp_module()

    if dtype is None:
        if image.dtype.name == 'uint8' or \
                image.dtype.name == 'uint16' or \
                image.dtype.name == 'int8' or \
                image.dtype.name == 'uint16' or \
                image.dtype.name == 'float16':
            dtype = xp.float16
        elif image.dtype.name == 'uint32' or \
                image.dtype.name == 'int32' or \
                image.dtype.name == 'float32':
            dtype = xp.float32
        elif image.dtype.name == 'uint64' or \
                image.dtype.name == 'int64' or \
                image.dtype.name == 'float64':
            dtype = xp.float32

    if type(Backend.current()) is NumpyBackend:
        dtype = numpy.float32

    original_dtype = image.dtype

    if minmax is None:

        min_value, max_value = 0, 0
        if quantile > 0:
            min_value = xp.percentile(image, q=100 * quantile, interpolation='lower')
            max_value = xp.percentile(image, q=100 * (1 - quantile), interpolation='higher')

        # if we did not use quantiles or we get some weird result, let's fallback to standard min max:
        if min_value >= max_value:
            min_value = xp.min(image)
            max_value = xp.max(image)

    else:
        min_value, max_value = minmax

    # Normalise:
    norm_alpha = ((high - low) / (max_value - min_value))
    norm_beta = (low - norm_alpha * min_value)

    # Ensure correct type:
    norm_alpha = xp.asarray(norm_alpha, dtype=dtype)
    norm_beta = xp.asarray(norm_beta, dtype=dtype)

    def _norm_function(_image):
        if not do_normalise:
            return _image
        _image = Backend.to_backend(_image, dtype=dtype)
        _image = element_wise_affine(_image, norm_alpha, norm_beta, out=_image if in_place else None)
        if clip:
            _image = _image.clip(low, high, out=_image)
        _image = _image.astype(dtype=dtype, copy=False)
        return _image

    # Denormalise:
    denorm_alpha = ((max_value - min_value) / (high - low))
    denorm_beta = (min_value - denorm_alpha * low)

    # Ensure correct type:
    denorm_alpha = xp.asarray(denorm_alpha, dtype=dtype)
    denorm_beta = xp.asarray(denorm_beta, dtype=dtype)

    def _denorm_function(_image):
        if not do_normalise:
            return _image
        _image = Backend.to_backend(_image, dtype=dtype)
        _image = element_wise_affine(_image, denorm_alpha, denorm_beta, out=_image if in_place else None)
        if clip:
            _image = _image.clip(min_value, max_value, out=_image)
        _image = _image.astype(dtype=original_dtype, copy=False)
        return _image

    return _norm_function, _denorm_function



#!/usr/bin/env python

"""

compatible with OpenGL, i.e.

mat4_perspective = gluPerspective
...etc


author: Martin Weigert
email: mweigert@mpi-cbg.de
"""
import numpy as np

from dexp.volumerender.quaternion import Quaternion


def mat4_scale(x=1., y=1., z=1.):
    return np.array([x, 0.0, 0.0, 0.,
                     0.0, y, 0.0, 0.0,
                     0.0, 0.0, z, 0.0,
                     0, 0, 0, 1.0], np.float32).reshape(4, 4)


def mat4_rotation(w=0, x=1, y=0, z=0):
    n = np.array([x, y, z], np.float32)
    n *= 1. / np.sqrt(1. * np.sum(n ** 2))
    q = Quaternion(np.cos(.5 * w), *(np.sin(.5 * w) * n))
    return q.toRotation4()


def mat4_rotation_euler(yaw=0, pitch=0, roll=0):
    """ in z-y'-x' convention


    """
    Z = mat4_rotation(yaw, 0, 0, 1)
    Y = mat4_rotation(pitch, 0, 1, 0)
    X = mat4_rotation(roll, 1, 0, 0)

    return np.dot(Z, np.dot(Y, X))


def mat4_perspective(fovy=45, aspect=1.,
                     z1=0.1, z2=10):
    """ like gluPerspective(fovy, aspect, zNear, zFar)
        fovy in degrees
    """
    f = 1. / np.tan(fovy / 180. * np.pi / 2.)
    return np.array([[1. * f / aspect, 0, 0, 0],
                     [0, f, 0, 0],
                     [0, 0, -1. * (z2 + z1) / (z2 - z1), -2. * z1 * z2 / (z2 - z1)],
                     [0, 0, -1, 0]], np.float32)


def mat4_stereo_perspective(fovy=45, aspect=1.,
                            z1=0.1, z2=10, eye_shift=0):
    h = z1 * np.tan(fovy / 180. * np.pi / 2.)
    w = h * aspect;

    return mat4_frustrum(-w - eye_shift, w - eye_shift, -h, h, z1, z2)


def mat4_frustrum(left, right, bottom, top, zNear, zFar):
    return np.array([[2. * zNear / (right - left), 0, 1. * (right + left) / (right - left), 0],
                     [0, 2. * zNear / (top - bottom), (top + bottom) / (top - bottom), 0],
                     [0, 0, -1. * (zFar + zNear) / (zFar - zNear), -2. * zFar * zNear / (zFar - zNear)],
                     [0, 0, -1., 0]], np.float32)


def mat4_ortho(x1=-1, x2=1,
               y1=-1, y2=1,
               z1=-1, z2=1):
    """ like glOrtho """
    ax, bx = x2 + x1, x2 - x1
    ay, by = y2 + y1, y2 - y1
    az, bz = z2 + z1, z2 - z1
    return np.array([[2. / bx, 0, 0, -1. * ax / bx],
                     [0, 2. / by, 0, -1. * ay / by],
                     [0, 0, -2. / bz, -1. * az / bz],
                     [0, 0, 0, 1.]], np.float32)


def mat4_identity():
    return np.identity(4)


def mat4_translate(x=0, y=0, z=0):
    M = mat4_identity()
    M[:3, 3] = x, y, z
    return M


def mat4_lookat(eye, center, up):
    _eye = np.array(eye, np.float32)
    _center = np.array(center, np.float32)
    _up = np.array(up, np.float32)

    _fwd = _center - _eye

    # normalize
    _fwd *= 1. / np.sqrt(np.sum(_fwd ** 2))

    s = np.cross(_fwd, _up)

    s *= 1. / np.sqrt(np.sum(s ** 2))

    _up = np.cross(s, _fwd)

    M = np.identity(4)
    M[0, :3] = s
    M[1, :3] = _up
    M[2, :3] = - _fwd

    return np.dot(M, mat4_translate(*(-_eye)))


# gluLookAt(GLdouble eyex, GLdouble eyey, GLdouble eyez, GLdouble centerx,
#       GLdouble centery, GLdouble centerz, GLdouble upx, GLdouble upy,
#       GLdouble upz)
# {
#     float forward[3], side[3], up[3];
#     GLfloat m[4][4];

#     forward[0] = centerx - eyex;
#     forward[1] = centery - eyey;
#     forward[2] = centerz - eyez;

#     up[0] = upx;
#     up[1] = upy;
#     up[2] = upz;

#     normalize(forward);

#     /* Side = forward x up */
#     cross(forward, up, side);
#     normalize(side);

#     /* Recompute up as: up = side x forward */
#     cross(side, forward, up);

#     __gluMakeIdentityf(&m[0][0]);
#     m[0][0] = side[0];
#     m[1][0] = side[1];
#     m[2][0] = side[2];

#     m[0][1] = up[0];
#     m[1][1] = up[1];
#     m[2][1] = up[2];

#     m[0][2] = -forward[0];
#     m[1][2] = -forward[1];
#     m[2][2] = -forward[2];

#     glMultMatrixf(&m[0][0]);
#     glTranslated(-eyex, -eyey, -eyez);
# }


if __name__ == '__main__':
    print(mat4_lookat([0, 0, 10], [0, 0, 0], [0, 1, 0]))

    # print rotMat(.1,0,0,1)
    # orthoM = projMatOrtho(-2,2,-2,2,-10,10)

    # perspM = projMatPerspective(45,1,.1,10)

    # x  = np.dot(orthoM,[0,0,0,1])
    # print x

    # x  = np.dot(perspM,[0,0,.1,1])
    # print x

    # print orthoM



from typing import Union, Tuple, Sequence

from dexp.processing.backends.backend import Backend
from dexp.processing.color.blend import blend_color_images
from dexp.processing.color.cairo_utils import get_array_for_cairo_surface


def insert_time_stamp(image: 'Array',
                      time_point_index: int,
                      nb_time_points: int,
                      start_time: float = 0,
                      time_interval: float = 1,
                      margin: float = 1,
                      translation: Union[str, Tuple[Union[int, float], ...]] = 'top_right',
                      color: Tuple[float, float, float, float] = None,
                      number_format: str = '{:.1f}',
                      font_name: str = "Helvetica",
                      font_size: float = 32,
                      unit: str = 's',
                      mode: str = 'max',
                      alpha: float = 1
                      ):
    """
    Inserts a scale bar into an image.

    Parameters
    ----------
    image: Image into which to insert the time stamp.
    time_point_index: Index (starting at 0 for start time) of the image.
    nb_time_points: Total number of time points.
    start_time: Start time for time stamp
    time_interval: Time interval inn units of time between consecutive images.
    margin: margin around bar expressed in units relative to the text height
    translation: Positions of the time stamp in pixels in numpy order: (y, x). Can also be a string: 'bottom_left', 'bottom_right', 'top_left', 'top_right'.
    color: Color of the bar and text as tuple of 4 values: (R, G, B, A)
    number_format: Format string to represent the start and end values.
    font_name: Font name.
    font_size: Font size in pixels.
    unit: Unit name.
    mode: Blending mode. See function 'blend_color_images' for available blending modes.
    alpha: Inset transparency.

    Returns
    -------
    Image with inserted time stamp.

    """

    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if color is None:
        color = (1, 1, 1, 1)

    # verify that the images are 2D RGB(A) images:
    if image.ndim != 3 or not image.shape[2] in (3, 4):
        raise ValueError("Image must be 2D RGB(A) images!")

    # Image shape:
    height, width = image.shape[:-1]

    # Dummy surface just for computing text width and height:
    import cairocffi as cairo
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
    context = cairo.Context(surface)
    context.scale(1, 1)

    # Configure text rendering:
    context.select_font_face(font_name,
                             cairo.FONT_SLANT_NORMAL,
                             cairo.FONT_WEIGHT_NORMAL)
    context.set_font_size(font_size)

    # longest text to be rendered:
    end_time = start_time + nb_time_points * time_interval
    time_text = number_format.format(end_time)
    text = f"{time_text} {unit}"

    # determine text height and width:
    text_width = context.text_extents(text)[2]
    text_height = context.text_extents(text)[3]
    end_time_text_width = context.text_extents(time_text + 'x')[2]

    # Margins:
    margin_height = margin * text_height
    margin_width = margin * text_height

    # Set the position of the time stamp
    if type(translation) == str:
        if 'top' in translation:
            y = margin_height + text_height
        elif 'bottom' in translation:
            y = height - (margin_height + text_height)
        if 'left' in translation:
            x = margin_width
        elif 'right' in translation:
            x = width - (margin_width + text_width)
    elif type(translation) == tuple:
        y, x = translation
    else:
        raise ValueError(f"Invalid translation: {translation}")

    def generate_time_stamp(timepoint: int):
        # Create surface:
        import cairocffi as cairo
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        context = cairo.Context(surface)
        context.scale(1, 1)

        # Configure text rendering:
        context.select_font_face(font_name,
                                 cairo.FONT_SLANT_NORMAL,
                                 cairo.FONT_WEIGHT_NORMAL)
        context.set_font_size(font_size)

        # First we turn off antialiasing, works better:
        context.set_antialias(cairo.ANTIALIAS_NONE)

        # Turn off alpha blending and clears the image:
        context.set_operator(cairo.OPERATOR_SOURCE)
        context.set_source_rgba(0, 0, 0, 0)
        context.rectangle(0, 0, 1, 1)
        context.fill()

        # Turn on antialiasing for text:
        context.set_antialias(cairo.ANTIALIAS_SUBPIXEL)

        # Turn back on alpha blending:
        context.set_operator(cairo.OPERATOR_OVER)

        # Set text color:
        context.set_source_rgba(*color)

        # Text to be rendered:
        time = start_time + timepoint * time_interval
        time_text = f"{number_format.format(time)}"
        time_width = context.text_extents(time_text + 'x')[2]
        unit_text = unit

        # draw text at correct location:
        context.move_to(x + end_time_text_width - time_width, y)
        context.show_text(time_text)
        context.move_to(x + end_time_text_width, y)
        context.show_text(unit_text)

        # Get array from surface:
        time_stamp = get_array_for_cairo_surface(surface)

        return time_stamp

    # generate time stamp:
    time_stamp = generate_time_stamp(time_point_index)

    # Blend images:
    image_with_time_stamp = blend_color_images(images=(image, time_stamp),
                                               alphas=(1, alpha),
                                               modes=('max', mode))

    return image_with_time_stamp



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.synthetic_datasets.multiview_data import generate_fusion_test_data
from dexp.processing.utils.blend_images import blend_images


def test_blend_numpy():
    with NumpyBackend():
        _test_blend()


def test_blend_cupy():
    try:
        with CupyBackend():
            _test_blend()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_blend(length_xy=128):
    xp = Backend.get_xp_module()

    image_gt, image_lowq, blend_a, blend_b, image1, image2 = generate_fusion_test_data(add_noise=False,
                                                                                       length_xy=length_xy,
                                                                                       length_z_factor=4,
                                                                                       dtype=numpy.float32)

    blended = blend_images(image1, image2, blend_a)

    assert blended is not image1
    assert blended is not image2
    assert blended.shape == image1.shape
    assert blended.shape == image2.shape
    assert blended.shape == blend_a.shape

    image_gt = Backend.to_numpy(image_gt)
    blended = Backend.to_numpy(blended)
    error = numpy.median(numpy.abs(image_gt - blended))
    print(f"error={error}")
    assert error < 23

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     viewer = Viewer()
    #     viewer.add_image(image_gt, name='image_gt')
    #     viewer.add_image(image1, name='image1')
    #     viewer.add_image(image2, name='image2')
    #     viewer.add_image(blend_a, name='blend_a')
    #     viewer.add_image(blended, name='blended')



from arbol import asection, aprint

from dexp.processing.backends._cupy.texture.texture import create_cuda_texture
from dexp.processing.backends.cupy_backend import CupyBackend


def test_cupy_texture_4channels():
    try:
        import cupy
        with CupyBackend():
            source = r'''
                extern "C"{
                __global__ void copyKernel(float* output,
                                           cudaTextureObject_t texObj,
                                           int width, int height)
                {
                    unsigned int x = blockIdx.x * blockDim.x + threadIdx.x;
                    unsigned int y = blockIdx.y * blockDim.y + threadIdx.y;
    
                    // Read from texture and write to global memory
                    float u = x+0.5f;
                    float v = y+0.5f;
                    if (x < width && y < height)
                    {
                        output[y * 4 *width + 4 *x +0] = tex2D<float4>(texObj, u, v).x;
                        output[y * 4 *width + 4 *x +1] = tex2D<float4>(texObj, u, v).y;
                        output[y * 4 *width + 4 *x +2] = tex2D<float4>(texObj, u, v).z;
                        output[y * 4 *width + 4 *x +3] = tex2D<float4>(texObj, u, v).w;
                    }
                }
                }
                '''
            width = 3
            height = 5

            # allocate input/output arrays
            tex_data = cupy.arange(width * height * 4, dtype=cupy.float32).reshape(height, width, 4)

            # set up a texture object
            texobj, cuda_array = create_cuda_texture(tex_data,
                                                     num_channels=4,
                                                     sampling_mode='nearest',
                                                     dtype=cupy.float32)

            real_output = cupy.zeros_like(tex_data)
            expected_output = tex_data.copy()

            # get the kernel, which copies from texture memory
            kernel = cupy.RawKernel(source, 'copyKernel')

            # launch it
            block_x = 4
            block_y = 4
            grid_x = (width + block_x - 1) // block_x
            grid_y = (height + block_y - 1) // block_y
            kernel((grid_x, grid_y), (block_x, block_y), (real_output, texobj, width, height))

            del texobj, cuda_array

            # test outcome
            assert cupy.allclose(real_output, expected_output)
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def test_cupy_texture_1channel_normcoord():
    try:
        import cupy
        with CupyBackend():
            source = r'''
                
                extern "C"{
                __global__ void texture_1channel_normcoord_kernel(float* output,
                                           cudaTextureObject_t texObj,
                                           int width, 
                                           int height)
                {
                    unsigned int x = blockIdx.x * blockDim.x + threadIdx.x;
                    unsigned int y = blockIdx.y * blockDim.y + threadIdx.y;
    
                    // Read from texture and write to global memory
                    float u = (float(x)+0.5f)/width;
                    float v = (float(y)+0.5f)/height;
                    
                    if (x < width && y < height)
                    {
                        float value = tex2D<float>(texObj, u, v);
                        printf("(%f, %f)=%f\n", u, v, value);
                        output[y * width + x] = value;
                    }
                }
                }
                '''
            width = 3
            height = 5

            # allocate input/output arrays
            tex_data = cupy.arange(width * height, dtype=cupy.float32).reshape(height, width)

            # set up a texture object
            texobj, cuda_array = create_cuda_texture(tex_data,
                                                     num_channels=1,
                                                     normalised_coords=True,
                                                     sampling_mode='linear',
                                                     dtype=cupy.float32)

            real_output = cupy.zeros_like(tex_data)
            expected_output = tex_data.copy()

            # get the kernel, which copies from texture memory
            kernel = cupy.RawKernel(source, 'texture_1channel_normcoord_kernel')

            # launch it
            block_x = 4
            block_y = 4
            grid_x = (width + block_x - 1) // block_x
            grid_y = (height + block_y - 1) // block_y
            kernel((grid_x, grid_y), (block_x, block_y), (real_output, texobj, width, height))

            del texobj, cuda_array

            # test outcome
            assert cupy.allclose(real_output, expected_output)
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def test_cupy_texture_1channel():
    try:
        import cupy
        with CupyBackend():
            source = r'''
                extern "C"{
                __global__ void copyKernel(float* output,
                                           cudaTextureObject_t texObj,
                                           int width, int height)
                {
                    unsigned int x = blockIdx.x * blockDim.x + threadIdx.x;
                    unsigned int y = blockIdx.y * blockDim.y + threadIdx.y;
    
                    // Read from texture and write to global memory
                    float u = x+0.5f;
                    float v = y+0.5f;
                    if (x < width && y < height)
                    {
                        float value = tex2D<float>(texObj, u, v);
                        printf("(%f, %f)=%f\n", u, v, value);
                        output[y * width + x] = value;
                    }
                }
                }
                '''
            width = 3
            height = 5

            # allocate input/output arrays
            tex_data = cupy.arange(width * height, dtype=cupy.float32).reshape(height, width)
            tex_data[1, 2] = 1

            # set up a texture object
            texobj, cuda_array = create_cuda_texture(tex_data,
                                                     num_channels=1,
                                                     sampling_mode='linear',
                                                     dtype=cupy.float32)

            real_output = cupy.zeros_like(tex_data)
            expected_output = tex_data.copy()

            # get the kernel, which copies from texture memory
            kernel = cupy.RawKernel(source, 'copyKernel')

            # launch it
            block_x = 4
            block_y = 4
            grid_x = (width + block_x - 1) // block_x
            grid_y = (height + block_y - 1) // block_y
            kernel((grid_x, grid_y), (block_x, block_y), (real_output, texobj, width, height))

            del texobj, cuda_array

            # test outcome
            assert cupy.allclose(real_output, expected_output)
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def test_basic_cupy_texture():
    try:
        import cupy
        with CupyBackend():
            source = r'''
                extern "C"{
                __global__ void copyKernel(float* output,
                                           cudaTextureObject_t texObj,
                                           int width, int height)
                {
                    unsigned int x = blockIdx.x * blockDim.x + threadIdx.x;
                    unsigned int y = blockIdx.y * blockDim.y + threadIdx.y;
    
                    // Read from texture and write to global memory
                    float u = x;
                    float v = y;
                    if (x < width && y < height)
                        output[y * width + x] = tex2D<float>(texObj, u, v);
                }
                }
                '''
            width = 8
            height = 16

            # set up a texture object
            ch = cupy.cuda.texture.ChannelFormatDescriptor(32, 0, 0, 0, cupy.cuda.runtime.cudaChannelFormatKindFloat)
            arr2 = cupy.cuda.texture.CUDAarray(ch, width, height)
            res = cupy.cuda.texture.ResourceDescriptor(cupy.cuda.runtime.cudaResourceTypeArray, cuArr=arr2)
            tex = cupy.cuda.texture.TextureDescriptor((cupy.cuda.runtime.cudaAddressModeClamp,
                                                       cupy.cuda.runtime.cudaAddressModeClamp),
                                                      cupy.cuda.runtime.cudaFilterModePoint,
                                                      cupy.cuda.runtime.cudaReadModeElementType)
            texobj = cupy.cuda.texture.TextureObject(res, tex)

            # allocate input/output arrays
            tex_data = cupy.arange(width * height, dtype=cupy.float32).reshape(height, width)
            real_output = cupy.zeros_like(tex_data)
            expected_output = cupy.zeros_like(tex_data)
            arr2.copy_from(tex_data)
            arr2.copy_to(expected_output)

            # get the kernel, which copies from texture memory
            ker = cupy.RawKernel(source, 'copyKernel')

            # launch it
            block_x = 4
            block_y = 4
            grid_x = (width + block_x - 1) // block_x
            grid_y = (height + block_y - 1) // block_y
            ker((grid_x, grid_y), (block_x, block_y), (real_output, texobj, width, height))

            del texobj, arr2

            # test outcome
            assert cupy.allclose(real_output, expected_output)
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def test_basic_cupy_texture_leak():
    try:
        import cupy
        with CupyBackend():
            # allocate input/output arrays
            length = 512
            tex_data = cupy.arange(length ** 3, dtype=cupy.float32).reshape(length, length, length)

            with asection("loop"):
                for i in range(100):
                    aprint(f"i={i}")
                    texobj, cuda_array = create_cuda_texture(tex_data,
                                                             num_channels=1,
                                                             sampling_mode='linear',
                                                             dtype=cupy.float32)

    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")



from arbol import asection

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.color.colormap import rgb_colormap
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data


def demo_colormap_numpy():
    with NumpyBackend():
        demo_colormap()


def demo_colormap_cupy():
    try:
        with CupyBackend():
            demo_colormap(length_xy=512)
        return True
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")
        return False


def demo_colormap(length_xy=120, display=True):
    with asection("generate data"):
        _, _, image = generate_nuclei_background_data(add_noise=False,
                                                      length_xy=length_xy,
                                                      length_z_factor=1,
                                                      background_stength=0.001,
                                                      sphere=True,
                                                      zoom=2)

        image -= image.min()
        image /= image.max()

    with asection("rgb_image"):
        rgb_image = rgb_colormap(image,
                                 cmap='turbo')

    if display:
        from napari import Viewer, gui_qt
        with gui_qt():
            def _c(array):
                return Backend.to_numpy(array)

            viewer = Viewer()
            viewer.add_image(_c(image), name='image')
            viewer.add_image(_c(rgb_image), name='rgb_image')


if __name__ == "__main__":
    if not demo_colormap_cupy():
        demo_colormap_numpy()



import threading
from abc import ABC, abstractmethod
from typing import Any

import numpy


class Backend(ABC):
    """
    Some description.
    """

    def __init__(self):
        """ Instanciates an Image Processing backend

        """

    _local = threading.local()

    @staticmethod
    def reset():
        if hasattr(Backend._local, 'backend_stack'):
            backends = Backend._local.backend_stack.copy()
            backends.reverse()
            for backend in backends:
                backend.__exit__()
            Backend._local.backend_stack = []

    @staticmethod
    def current(raise_error_if_none: bool = False):

        if hasattr(Backend._local, 'backend_stack'):
            backend_stack = Backend._local.backend_stack
            if backend_stack is None or len(backend_stack) == 0:
                if raise_error_if_none:
                    raise RuntimeError("No backend available in current thread context")
                else:
                    # aprint("Warning: no backend available in current thread context! falling back to a numpy backend! ")
                    from dexp.processing.backends.numpy_backend import NumpyBackend
                    return NumpyBackend()
            backend = backend_stack[-1]
            return backend
        else:
            if raise_error_if_none:
                raise RuntimeError("No backend available in current thread context")
            else:
                # aprint("Warning: no backend available in current thread context! falling back to a numpy backend! ")
                from dexp.processing.backends.numpy_backend import NumpyBackend
                return NumpyBackend()

    @staticmethod
    def set(backend: 'Backend'):
        if not hasattr(Backend._local, 'backend_stack'):
            Backend._local.backend_stack = []
        Backend._local.backend_stack.append(backend)

    @staticmethod
    def to_numpy(array, dtype=None, force_copy: bool = False) -> numpy.ndarray:
        return Backend.current()._to_numpy(array, dtype=dtype, force_copy=force_copy)

    @staticmethod
    def to_backend(array, dtype=None, force_copy: bool = False) -> Any:
        return Backend.current()._to_backend(array, dtype=dtype, force_copy=force_copy)

    @staticmethod
    def get_xp_module(array=None) -> Any:
        return Backend.current()._get_xp_module(array)

    @staticmethod
    def get_sp_module(array=None) -> Any:
        return Backend.current()._get_sp_module(array)

    def __enter__(self):
        if not hasattr(Backend._local, 'backend_stack'):
            Backend._local.backend_stack = []
        Backend._local.backend_stack.append(self)
        return self

    def __exit__(self, type, value, traceback):
        Backend._local.backend_stack.pop()

    @abstractmethod
    def copy(self, *args, **kwargs):
        raise NotImplementedError('Method not implemented!')

    @abstractmethod
    def clear_memory_pool(self):
        import gc
        gc.collect()

    def close(self):
        """ Releases all ressources allocated/cached by backend (if can be done safely)

        """
        self.clear_memory_pool()

    @abstractmethod
    def _to_numpy(self, array, dtype=None, force_copy: bool = False) -> numpy.ndarray:
        """ Converts backend array to numpy. If array is already a numpy array it is returned unchanged.

        Parameters
        ----------
        array : backend array to be converted
        dtype : coerce array to given dtype
        force_copy : forces the return array to be a copy

        Returns
        -------
        array converted to backend

        """
        raise NotImplementedError('Method not implemented!')

    @abstractmethod
    def _to_backend(self, array, dtype=None, force_copy: bool = False) -> Any:
        """ Converts numpy array to backend array, if already backend array, then it is returned unchanged

        Parameters
        ----------
        array : numpy array to be converted
        dtype : coerce array to given dtype
        force_copy : forces the return array to be a copy
        """
        raise NotImplementedError('Method not implemented!')

    @abstractmethod
    def _get_xp_module(self, array=None) -> Any:
        """ Returns the numpy-like module for a given array

        Parameters
        ----------
        array : array from which to get the numpy-like module


        Returns
        -------
        numpy-like module
        """
        raise NotImplementedError('Method not implemented!')

    @abstractmethod
    def _get_sp_module(self, array=None) -> Any:
        """ Returns the scipy-like module for a given array

        Parameters
        ----------
        array : array from which to get the scipy-like module

        Returns
        -------
        scipy-like module
        """
        raise NotImplementedError('Method not implemented!')



from arbol import asection
from skimage.color import gray2rgba
from skimage.data import camera

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.color.border import add_border


def demo_border_numpy():
    with NumpyBackend():
        demo_border()


def demo_border_cupy():
    try:
        with CupyBackend():
            demo_border()
        return True
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")
        return False


def demo_border(display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    with asection("generate data"):
        image = Backend.to_backend(gray2rgba(camera()))

    with asection("Add border..."):
        image_with_border = add_border(image,
                                       width=3,
                                       color=(1, 1, 1, 1)
                                       )

    with asection("Overlay border..."):
        image_with_border_overlayed = add_border(image,
                                                 width=3,
                                                 color=(1, 1, 1, 1),
                                                 over_image=True,
                                                 )

    if display:
        from napari import Viewer, gui_qt
        with gui_qt():
            def _c(array):
                return Backend.to_numpy(array)

            viewer = Viewer()
            viewer.add_image(_c(image), name='image', rgb=True)
            viewer.add_image(_c(image_with_border), name='image_with_border', rgb=True)
            viewer.add_image(_c(image_with_border_overlayed), name='image_with_border_overlayed', rgb=True)
            viewer.grid.enabled = True


if __name__ == "__main__":
    if not demo_border_cupy():
        demo_border_numpy()



from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.color.demo.demo_border import demo_border


def test_border_numpy():
    with NumpyBackend():
        _test_border()


def test_border_cupy():
    try:
        with CupyBackend():
            _test_border()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_border():
    demo_border(display=False)



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.utils.scatter_gather_i2v import scatter_gather_i2v
from dexp.utils.timeit import timeit


def test_scatter_gather_i2v_numpy():
    with NumpyBackend():
        _test_scatter_gather_i2v()


def test_scatter_gather_i2v_cupy():
    try:
        with CupyBackend():
            _test_scatter_gather_i2v(length_xy=512, splits=4, filter_size=7)
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_scatter_gather_i2v(ndim=3, length_xy=128, splits=4, filter_size=7):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    image1 = numpy.random.uniform(0, 1, size=(length_xy,) * ndim)
    image2 = numpy.random.uniform(0, 1, size=(length_xy,) * ndim)

    def f(x, y):
        return (xp.asarray([x.min(), x.max()]), xp.asarray([y.max(), y.mean(), y.min()]))

    try:
        with timeit("f"):
            result_ref_1, result_ref_2 = Backend.to_numpy(f(Backend.to_backend(image1), Backend.to_backend(image2)))
    except:
        print("Can't run this, not enough GPU memory!")
        result_ref_1 = 0
        result_ref_2 = 0

    with timeit("scatter_gather(f)"):
        chunks = (length_xy // splits,) * ndim
        result1, result2 = scatter_gather_i2v(f, (image1, image2), chunks=chunks, margins=8)

    print(result1.shape)
    print(result2.shape)

    assert result1.ndim == ndim + 1
    assert result2.ndim == ndim + 1

    assert result1.shape[-1] == 2
    assert result2.shape[-1] == 3

    mean = result1.mean(axis=tuple(a for a in range(ndim)))
    result1 -= mean
    result1 = Backend.to_numpy(result1)
    error = numpy.linalg.norm(result1.ravel(), ord=1) / result1.size
    print(f"Error = {error}")
    assert error < 0.001

    mean = result2.mean(axis=tuple(a for a in range(ndim)))
    result2 -= mean
    result2 = Backend.to_numpy(result2)
    error = numpy.linalg.norm(result2.ravel(), ord=1) / result2.size
    print(f"Error = {error}")
    assert error < 0.001

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     viewer = Viewer()
    #     viewer.add_image(result, name='result', color=False)
    #



import numpy
from arbol import aprint

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend


def equalise_intensity(image1,
                       image2,
                       zero_level=90,
                       quantile_low=0.01,
                       quantile_high=0.99,
                       project_axis: int = 0,
                       max_voxels: int = 1e7,
                       correction_ratio: float = None,
                       copy: bool = True,
                       internal_dtype=None):
    """
    Equalise intensity between two images

    Parameters
    ----------
    image1 : first image to equalise
    image2 : second image to equalise
    zero_level : zero level -- removes this value if that's the minimal voxel value expected for both images
    quantile_low, quantile_high : quantile for computing the robust min and max values in image
    project_axis : Axis over which to project image to speed up computation
    max_voxels : maximal number of voxels to use to compute min and max values.
    correction_ratio : If provided, this value is used instead of calculating the equalisation value based on the two images.
    This is the value that should be multiplied to the second image (image2) so as to equalise it to teh first image (image1).
    copy : Set to True to force copy of images.
    internal_dtype : dtype to use internally for computation.

    Returns
    -------
    The two arrays intensity equalised.

    """
    xp = Backend.get_xp_module()

    # Sanity checks:
    if image1.dtype != image2.dtype:
        raise ValueError("Two images must have same dtype!")

    if internal_dtype is None:
        internal_dtype = image1.dtype

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = numpy.float32

    # Save original dtype:
    original_dtype = image1.dtype

    # Move images to backend.
    image1 = Backend.to_backend(image1, dtype=internal_dtype, force_copy=copy)
    image2 = Backend.to_backend(image2, dtype=internal_dtype, force_copy=copy)

    # If not provided, compute correction ratio:
    if correction_ratio is None:

        # To reduce computation effort, we operate on the max projected images:
        proj_image1 = xp.max(image1, axis=project_axis)
        proj_image2 = xp.max(image2, axis=project_axis)

        # instead of computing the rations on all voxels, we reduce to a dense subset:
        reduction = max(1, 4 * (int(proj_image1.size / max_voxels) // 4))
        strided_image1 = proj_image1.ravel()[::reduction].astype(numpy.float32, copy=False)
        strided_image2 = proj_image2.ravel()[::reduction].astype(numpy.float32, copy=False)

        # We determine a robust maximum for voxel intensities:
        highvalue1 = xp.percentile(strided_image1, q=quantile_high * 100, interpolation='higher')
        highvalue2 = xp.percentile(strided_image2, q=quantile_high * 100, interpolation='higher')

        # we set a 'high-value' threshold to the minimum of both robust maximums:
        threshold = min(highvalue1, highvalue2)

        # We find the voxels that are above that threshold in both images:
        mask1 = strided_image1 >= threshold
        mask2 = strided_image2 >= threshold
        mask = xp.logical_and(mask1, mask2)

        # And extract these values from both images:
        highvalues1 = strided_image1[mask]
        highvalues2 = strided_image2[mask]

        # And a robust minimum:
        lowvalue1 = xp.percentile(strided_image1, q=quantile_low * 100)
        lowvalue2 = xp.percentile(strided_image2, q=quantile_low * 100)

        # Compute ratios:
        range1 = highvalues1 - lowvalue1
        range2 = highvalues2 - lowvalue2
        ratios = (range1 / range2)

        # Keep only valid ratios:
        valid = xp.logical_and(range1 != 0, range2 != 0)
        ratios = ratios[valid]

        # Free memory:
        del mask, mask1, mask2, highvalues1, highvalues2, strided_image1, strided_image2

        # Number of values in ratio:
        nb_values = ratios.size
        if nb_values < 16:
            raise ValueError(f"No enough values ({nb_values}<16) to compute correction ratio! Relax percentile or reduction! ")
        elif nb_values < 128:
            aprint(f"Warning: too few ratio values ({nb_values}<128) to compute correction ratio! Relax percentile or reduction!")

        correction_ratio = xp.percentile(ratios.astype(internal_dtype, copy=False), q=50)
    else:
        correction_ratio = Backend.to_backend(correction_ratio, dtype=internal_dtype)

    # remove zero level and clip:
    if zero_level != 0:
        image1 = xp.clip(image1, a_min=zero_level, a_max=None, out=image1)
        image1 -= zero_level
        image2 = xp.clip(image2, a_min=zero_level, a_max=None, out=image2)
        image2 -= zero_level

    # compute inverse ratios and cast to internal type:
    inverse_correction_ratio = 1 / correction_ratio
    correction_ratio = correction_ratio.astype(dtype=internal_dtype)
    inverse_correction_ratio = inverse_correction_ratio.astype(dtype=internal_dtype)

    # Apply correction ratio while only increasing intensity of the dimmer image:
    if correction_ratio > 1:
        image2 *= correction_ratio
    else:
        image1 *= inverse_correction_ratio

    # cast back to original dtype:
    image1 = image1.astype(original_dtype, copy=False)
    image2 = image2.astype(original_dtype, copy=False)

    # KEEP FOR DEBUGGING:
    # from napari import Viewer
    # import napari
    # with napari.gui_qt():
    #     def _c(array):
    #         return backend.to_numpy(array)
    #
    #     viewer = Viewer()
    #     viewer.add_image(_c(image1), name='image_1')
    #     viewer.add_image(_c(image1), name='image_2')

    return image1, image2, correction_ratio



import numpy
from arbol import asection, aprint

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data
from dexp.processing.utils.center_of_mass import center_of_mass


def test_com_numpy():
    with NumpyBackend():
        _test_com()


def test_com_cupy():
    try:
        with CupyBackend():
            _test_com()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_com(length_xy=128):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    with asection("generate data"):
        _, _, image = generate_nuclei_background_data(add_noise=False,
                                                      length_xy=length_xy,
                                                      length_z_factor=1,
                                                      background_stength=0.001,
                                                      sphere=True,
                                                      radius=0.5,
                                                      zoom=2,
                                                      add_offset=False,
                                                      dtype=numpy.uint16)

        # from napari import Viewer, gui_qt
        # with gui_qt():
        #     def _c(array):
        #         return Backend.to_numpy(array)
        #     viewer = Viewer()
        #     viewer.add_image(_c(image), name='image')

        com_before = center_of_mass(image)
        aprint(f"com_before: {com_before}")

        image_shifted = sp.ndimage.shift(image, shift=(50, 70, -23), order=1, mode='nearest')

        com_after = center_of_mass(image_shifted)
        aprint(f"com_after: {com_after}")

        com_after_bb = center_of_mass(image_shifted, offset_mode='p=75', bounding_box=True)
        aprint(f"com_after_bb: {com_after_bb}")

        # from napari import Viewer, gui_qt
        # with gui_qt():
        #     def _c(array):
        #         return Backend.to_numpy(array)
        #     viewer = Viewer()
        #     viewer.add_image(_c(image), name='image')
        #     viewer.add_image(_c(image_shifted), name='image_shifted')

        assert xp.mean(xp.absolute((com_after - xp.asarray((50, 70, -23)) - com_before))) < 10
        assert xp.mean(xp.absolute((com_after_bb - xp.asarray((50, 70, -23)) - com_before))) < 10



from typing import List, Sequence

import numpy
from arbol import aprint, asection
from dask.array import Array

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.registration.model.sequence_registration_model import SequenceRegistrationModel
from dexp.processing.registration.model.translation_registration_model import TranslationRegistrationModel
from dexp.processing.registration.translation_nd_proj import register_translation_proj_nd
from dexp.processing.utils.center_of_mass import center_of_mass
from dexp.processing.utils.linear_solver import linsolve


def image_stabilisation(image: 'Array',
                        axis: int = 0,
                        internal_dtype=None,
                        **kwargs
                        ) -> SequenceRegistrationModel:
    """
    Computes a sequence stabilisation model for an image sequence indexed along a specified axis.


    Parameters
    ----------
    image: image to stabilise
    axis: sequence axis along which to stabilise image
    internal_dtype : internal dtype for computation
    **kwargs: argument passthrough to the pairwise registration method.

    Returns
    -------
    Sequence registration model

    """
    xp = Backend.get_xp_module()

    image = Backend.to_backend(image, dtype=internal_dtype)

    length = image.shape[axis]
    image_sequence = list(xp.take(image, axis=axis, indices=range(0, length)))

    return image_sequence_stabilisation(image_sequence=image_sequence,
                                        internal_dtype=internal_dtype,
                                        **kwargs)


def image_sequence_stabilisation(image_sequence: Sequence['Array'],
                                 preload_images: bool = True,
                                 mode: str = 'translation',
                                 max_range: int = 7,
                                 min_confidence: float = 0.5,
                                 enable_com: bool = False,
                                 quantile: float = 0.5,
                                 bounding_box: bool = False,
                                 tolerance: float = 1e-7,
                                 order_error: float = 2.0,
                                 order_reg: float = 1.0,
                                 alpha_reg: float = 0.1,
                                 debug_output: str = None,
                                 internal_dtype=None,
                                 **kwargs
                                 ) -> SequenceRegistrationModel:
    """
    Computes a sequence stabilisation model for an image sequence.


    Parameters
    ----------
    image: image to stabilise
    axis: sequence axis along which to stabilise image
    mode: registration mode. For now only 'translation' is available.
    max_range: maximal distance, in time points, between pairs of images to registrate.
    min_confidence: minimal confidence to accept a pairwise registration
    enable_com: enable center of mass fallback when standard registration fails.
    quantile: quantile to cut-off background in center-of-mass calculation
    bounding_box: if True, the center of mass of the bounding box of non-zero pixels is returned.
    tolerance: tolerance for linear solver.
    order_error: order for linear solver error term.
    order_reg: order for linear solver regularisation term.
    alpha_reg: multiplicative coefficient for regularisation term.
    internal_dtype : internal dtype for computation
    **kwargs: argument passthrough to the pairwise registration method, see 'register_translation_nd'.

    Returns
    -------
    Sequence registration model

    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    image_sequence = list(image_sequence)

    if preload_images:
        with asection(f"Preloading images to backend..."):
            image_sequence = list(Backend.to_backend(image) for image in image_sequence)

    length = len(image_sequence)

    scales = list(i for i in range(max_range) if i < length)

    aprint(f"Scales: {scales}")

    with asection(f"Registering image sequence of length: {length}"):

        ndim = image_sequence[0].ndim

        uv_set = set()
        with asection(f"Enumerating pairwise registrations needed..."):

            for scale_index, scale in enumerate(scales):
                for offset in range(0, scale):
                    for u in range(offset, length, scale):

                        v = u + scale

                        # if u < 0:
                        #     u = 0
                        if v >= length:
                            continue

                        atuple = tuple((u, v))
                        # aprint(f"Pair: ({u}, {v}) added")
                        uv_set.add(atuple)

        pairwise_models = []
        with asection(f"Computing pairwise registrations for {len(uv_set)} (u,v) pairs..."):
            for u, v in uv_set:
                image_u = image_sequence[u]
                image_v = image_sequence[v]
                model = _pairwise_registration(u, v,
                                               image_u, image_v,
                                               mode,
                                               min_confidence,
                                               enable_com,
                                               quantile,
                                               bounding_box,
                                               internal_dtype,
                                               **kwargs)
                if model is not None:
                    pairwise_models.append(model)

        nb_models = len(pairwise_models)
        aprint(f"Number of models obtained: {nb_models} for a sequence of length:{length}")

        if debug_output is not None:
            with asection(f"Generating pairwise registration confidence matrix: '{debug_output}' "):
                import matplotlib.pyplot as plt
                import seaborn as sns
                sns.set_theme()
                plt.clf()
                plt.cla()

                array = numpy.zeros((length, length))
                for tp, model in enumerate(pairwise_models):
                    u = model.u
                    v = model.v
                    array[u, v] = model.overall_confidence()
                sns.heatmap(array)

                plt.savefig(debug_output + '_prcm.pdf')

        with asection(f"Solving for optimal sequence registration"):
            with NumpyBackend():
                xp = Backend.get_xp_module()
                sp = Backend.get_sp_module()

                if mode == 'translation':

                    # prepares list of models:
                    translation_models: List[TranslationRegistrationModel] = list(TranslationRegistrationModel(xp.zeros((ndim,), dtype=internal_dtype)) for _ in range(length))

                    # initialise count for average:
                    for model in translation_models:
                        model.count = 0

                    for d in range(ndim):

                        # instantiates system to solve with zeros:
                        a = xp.zeros((nb_models + 1, length), dtype=xp.float32)
                        y = xp.zeros((nb_models + 1,), dtype=xp.float32)

                        # iterates through all pairwise registrations:
                        zero_vector = None
                        for tp, model in enumerate(pairwise_models):
                            u = model.u
                            v = model.v
                            confidence = model.overall_confidence()

                            if tp == 0:
                                # we make sure that all shifts are relative to the first timepoint:
                                zero_vector = model.shift_vector[d].copy()

                            vector = model.shift_vector[d] - zero_vector

                            # Each pairwise registration defines a constraint that is added to the matrix:
                            y[tp] = vector.copy()
                            a[tp, u] = +1
                            a[tp, v] = -1

                            # For each time point we collect the average confidence of all the pairwise_registrations:
                            translation_models[u].confidence += confidence
                            translation_models[u].count += 1
                            translation_models[v].confidence += confidence
                            translation_models[v].count += 1

                        # Forces solution to have no displacement for first time point:
                        y[-1] = 0
                        a[-1, 0] = 1

                        # move matrix to sparse domain (much faster!):
                        a = sp.sparse.coo_matrix(a)

                        # solve system:
                        x_opt = linsolve(a, y,
                                         tolerance=tolerance,
                                         order_error=order_error,
                                         order_reg=order_reg,
                                         alpha_reg=alpha_reg)

                        # detrend:
                        # x_opt = sp.signal.detrend(x_opt)

                        # sets the shift vectors for the resulting sequence reg model, and compute average confidences:
                        for tp in range(length):
                            translation_models[tp].shift_vector[d] = -x_opt[tp]
                            translation_models[tp].confidence /= translation_models[tp].count

                    model = SequenceRegistrationModel(model_list=translation_models)

                else:
                    raise ValueError(f"Unsupported sequence stabilisation mode: {mode}")

            if debug_output is not None:
                with asection(f"Generating shift vector plots"):
                    model.plot(debug_output)

            return model

            # if not image_a.dtype == image_b.dtype:
            #     raise ValueError("Arrays must have the same dtype")
            #
            # if internal_dtype is None:
            #     internal_dtype = image_a.dtype
            #
            # if type(Backend.current()) is NumpyBackend:
            #     internal_dtype = xp.float32
            #
            # image_a = Backend.to_backend(image_a, dtype=internal_dtype)
            # image_b = Backend.to_backend(image_b, dtype=internal_dtype)


def _pairwise_registration(u, v,
                           image_u, image_v,
                           mode,
                           min_confidence,
                           enable_com,
                           quantile,
                           bounding_box,
                           internal_dtype,
                           **kwargs):
    image_u = Backend.to_backend(image_u, dtype=internal_dtype)
    image_v = Backend.to_backend(image_v, dtype=internal_dtype)

    if mode == 'translation':
        model = register_translation_proj_nd(image_u, image_v,
                                             _display_phase_correlation=False,
                                             **kwargs)
        model.u = u
        model.v = v
        confidence = model.overall_confidence()

        if (confidence < min_confidence):
            if enable_com:
                # aprint(f"Warning: low confidence ({confidence}) for pair: ({u}, {v}) falling back to center-of-mass calculation")
                offset_mode = f'p={quantile * 100}'
                com_u = center_of_mass(image_u, mode='full', projection_type='max-min', offset_mode=offset_mode, bounding_box=bounding_box)
                com_v = center_of_mass(image_v, mode='full', projection_type='max-min', offset_mode=offset_mode, bounding_box=bounding_box)
                model = TranslationRegistrationModel(shift_vector=com_u - com_v, confidence=min_confidence)
                model.u = u
                model.v = v
            else:
                model = None

    else:
        raise ValueError(f"Unsupported sequence stabilisation mode: {mode}")

    return model



from typing import Tuple, Union

import numpy

from dexp.processing.backends.backend import Backend


def butterworth_kernel(shape: Tuple[int, ...],
                       cutoffs: Union[float, Tuple[float, ...]],
                       cutoffs_in_freq_units=False,
                       epsilon: float = 1,
                       order: int = 3,
                       frequency_domain: bool = False,
                       dtype=numpy.float32):
    """

    Parameters
    ----------
    shape : filter shape
    cutoffs : Butterworth cutoffs.
    cutoffs_in_freq_units : If True, the cutoffs are specified in frequency units. If False, the units are in normalised within [0,1]
    epsilon : maximum cutoff gain
    order : Butterworth filter order
    frequency_domain : True to return the kernel in the frequency domain
    dtype : dtype to return kernel

    Returns
    -------
    Normalised butterworth filter kernel.

    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    ndim = len(shape)

    if type(cutoffs) is not tuple:
        cutoffs = (cutoffs,) * ndim

    if ndim == 1:

        lx, = shape
        cx, = cutoffs

        x = xp.fft.fftfreq(lx) if cutoffs_in_freq_units else xp.linspace(-1, 1, lx)

        # An array with every pixel = radius relative to center
        freq = xp.abs(x / cx)

    elif ndim == 2:

        ly, lx = shape
        cy, cx = cutoffs

        x = xp.fft.fftfreq(lx) if cutoffs_in_freq_units else xp.linspace(-1, 1, lx)
        y = xp.fft.fftfreq(ly) if cutoffs_in_freq_units else xp.linspace(-1, 1, ly)

        # An array with every pixel = radius relative to center
        freq = ((x / cx) ** 2)[xp.newaxis, xp.newaxis, :] + ((y / cy) ** 2)[xp.newaxis, :, xp.newaxis]

    elif ndim == 3:
        lz, ly, lx = shape
        cz, cy, cx = cutoffs

        x = xp.fft.fftfreq(lx) if cutoffs_in_freq_units else xp.linspace(-1, 1, lx)
        y = xp.fft.fftfreq(ly) if cutoffs_in_freq_units else xp.linspace(-1, 1, ly)
        z = xp.fft.fftfreq(lz) if cutoffs_in_freq_units else xp.linspace(-1, 1, lz)

        # An array with every pixel = radius relative to center
        freq = ((x / cx) ** 2)[xp.newaxis, xp.newaxis, :] + ((y / cy) ** 2)[xp.newaxis, :, xp.newaxis] + ((z / cz) ** 2)[:, xp.newaxis, xp.newaxis]

    kernel_fft = 1 / xp.sqrt(1.0 + (epsilon ** 2) * (freq ** order))

    kernel_fft = xp.squeeze(kernel_fft)

    if not cutoffs_in_freq_units:
        kernel_fft = sp.fft.ifftshift(kernel_fft)

    if frequency_domain:
        return kernel_fft
    else:
        kernel = sp.fft.fftshift(xp.real(sp.fft.ifftn(kernel_fft)))
        kernel = kernel / kernel.sum()
        kernel = kernel.astype(dtype, copy=False)
        return kernel



from arbol import aprint

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.registration.model.model_io import from_json, model_list_to_file, model_list_from_file
from dexp.processing.registration.model.sequence_registration_model import SequenceRegistrationModel
from dexp.processing.registration.model.translation_registration_model import TranslationRegistrationModel
from dexp.processing.registration.model.warp_registration_model import WarpRegistrationModel


def test_model_io_numpy():
    with NumpyBackend():
        _test_model_io_to_json()
        _test_model_io_to_file()


def test_model_io_cupy():
    try:
        with CupyBackend():
            _test_model_io_to_json()
            _test_model_io_to_file()
    except ModuleNotFoundError:
        print("Cupy module not found! demo ignored")


def _test_model_io_to_json():
    xp = Backend.get_xp_module()

    # Testing read/write translation model to json:
    model = TranslationRegistrationModel(shift_vector=[-1, -5, 13], confidence=0.6)
    json_str = model.to_json()
    new_model = from_json(json_str)
    assert (new_model.shift_vector == model.shift_vector).all()
    assert (new_model.confidence == model.confidence).all()

    # Testing read/write warp model to json:
    magnitude = 15
    warp_grid_size = 4
    vector_field = xp.random.uniform(low=-magnitude, high=+magnitude, size=(warp_grid_size,) * 3 + (3,))
    confidence = xp.random.uniform(low=-magnitude, high=+magnitude, size=(warp_grid_size,) * 3)
    model = WarpRegistrationModel(vector_field=vector_field, confidence=confidence)
    json_str = model.to_json()
    new_model = from_json(json_str)
    assert (new_model.vector_field == model.vector_field).all()
    assert (new_model.confidence == model.confidence).all()

    # Testing read/write translation sequence model to json:
    model_list = list(TranslationRegistrationModel(shift_vector=[-1 * (i / 100.0), -5 + 0.1 * i, 13 - 0.2 * i], confidence=0.6) for i in range(0, 10))
    model_list += list(WarpRegistrationModel(vector_field=xp.random.uniform(low=-magnitude, high=+magnitude, size=(warp_grid_size,) * 3 + (3,)),
                                             confidence=xp.random.uniform(low=-magnitude, high=+magnitude, size=(warp_grid_size,) * 3)) for _ in range(0, 10))
    model = SequenceRegistrationModel(model_list)
    json_str = model.to_json()
    new_model = from_json(json_str)
    assert model == new_model


def _test_model_io_to_file():
    xp = Backend.get_xp_module()

    magnitude = 15
    warp_grid_size = 4
    vector_field = xp.random.uniform(low=-magnitude, high=+magnitude, size=(warp_grid_size,) * 3 + (3,))
    confidence = xp.random.uniform(low=-magnitude, high=+magnitude, size=(warp_grid_size,) * 3)

    import tempfile
    with tempfile.NamedTemporaryFile() as tmp:
        aprint(f"Temp file: {tmp.name}")

        model_list_1 = list([TranslationRegistrationModel(shift_vector=[-1, i, 2 * i], confidence=0.6 - 0.01 * i) for i in range(3)])
        model_list_2 = list([WarpRegistrationModel(vector_field=vector_field * (1 + 0.1 * i), confidence=confidence / i) for i in range(3)])
        model_list = model_list_1 + model_list_2
        model_list_to_file(tmp.name, model_list)

        loaded_model_list = model_list_from_file(tmp.name)

        assert model_list == loaded_model_list



import re
from fnmatch import fnmatch
from os import listdir, cpu_count
from os.path import join
from typing import Sequence, Tuple, Any

import dask
import numcodecs
import numpy
from arbol.arbol import aprint
from cachey import Cache
from dask import array, delayed
from numpy import uint16, frombuffer

from dexp.datasets.base_dataset import BaseDataset

numcodecs.blosc.use_threads = True
numcodecs.blosc.set_nthreads(cpu_count() // 2)


class CCDataset(BaseDataset):

    def __init__(self, path, cache_size=8e9):

        super().__init__(dask_backed=False)

        self.folder = path
        self._channels = []
        self._index_files = {}

        all_files = list(listdir(path))
        # print(all_files)

        for file in all_files:
            if fnmatch(file, '*.index.txt'):
                if not file.startswith('._'):
                    channel = file.replace('.index.txt', '')
                    self._channels.append(channel)
                    self._index_files[channel] = join(path, file)

        # print(self._channels)
        # print(self._index_files)

        self._nb_time_points = {}
        self._times_sec = {}
        self._shapes = {}
        self._time_points = {}

        for channel in self._channels:
            self._parse_channel(channel)

        self.cache = Cache(cache_size)  # Leverage two gigabytes of memory

    def _parse_channel(self, channel):

        index_file = self._index_files[channel]

        with open(index_file) as f:
            lines = f.readlines()

        lines = [line.strip() for line in lines]

        lines = [re.split(r'\t+', line) for line in lines]

        self._time_points[channel] = []

        for line in lines:
            time_point = int(line[0])
            time_sec = float(line[1])
            shape = eval('(' + line[2] + ')')[::-1]

            self._times_sec[(channel, time_point)] = time_sec
            self._shapes[(channel, time_point)] = shape

            self._time_points[channel].append(time_point)

        self._nb_time_points[channel] = len(self._time_points[channel])

    def _get_stack_file_name(self, channel, time_point):

        return join(self.folder, 'stacks', channel, str(time_point).zfill(6) + '.raw')

    def _get_array_for_stack_file(self, file_name, shape=None):

        try:
            with open(file_name, 'rb') as my_file:
                aprint(f"Accessing file: {file_name}")
                buffer = my_file.read()

                dt = numpy.dtype(uint16)
                dt = dt.newbyteorder('L')

                array = frombuffer(buffer, dtype=dt)

            if not shape is None:
                array = array.reshape(shape)

            return array

        except FileNotFoundError:
            aprint(f"Could not find file: {file_name} for array of shape: {shape}")
            return numpy.zeros(shape, dtype=uint16)

    def _get_slice_array_for_stack_file_and_z(self, file_name, shape, z):

        try:
            with open(file_name, 'rb') as file:
                aprint(f"Accessing file: {file_name} at z={z}")

                length = shape[1] * shape[2] * numpy.dtype(uint16).itemsize
                offset = z * length

                file.seek(offset)

                buffer = file.read(length)
                array = frombuffer(buffer, dtype=uint16)

                array = array.reshape(shape[1:])
                return array

        except FileNotFoundError:
            aprint(f"Could  not find file: {file_name} for array of shape: {shape} at z={z}")
            return numpy.zeros(shape[1:], dtype=uint16)

    def close(self):
        # Nothing to do...
        pass

    def channels(self):
        return list(self._channels)

    def shape(self, channel: str, time_point: int = 0) -> Sequence[int]:
        if (channel, time_point) in self._shapes:
            return (self._nb_time_points[channel],) + self._shapes[(channel, time_point)]
        else:
            return ()

    def dtype(self, channel: str):
        return numpy.uint16

    def info(self, channel: str = None) -> str:
        if channel:
            info_str = f"Channel: '{channel}', nb time points: {self.shape(channel)[0]}, shape: {self.shape(channel)[1:]} "
            return info_str
        else:
            return self.tree()

    def get_metadata(self):
        # TODO: implement this!
        return {}

    def get_array(self, channel: str, per_z_slice: bool = True, wrap_with_dask: bool = False):

        if False:
            # For some reason this is slower, should have been faster!:
            # Construct a small Dask array for every lazy value:
            arrays = [self.get_stack(channel, time_point, per_z_slice) for time_point in self._time_points[channel]]
            stacked_array = array.stack(arrays, axis=0)  # Stack all small Dask arrays into one
            return stacked_array

        else:
            # Lazy and memorized version of get_stack:
            lazy_get_stack = delayed(self.cache.memoize(self.get_stack), pure=True)

            # Lazily load each stack for each time point:
            lazy_stacks = [lazy_get_stack(channel, time_point, per_z_slice) for time_point in self._time_points[channel]]

            # Construct a small Dask array for every lazy value:
            arrays = [array.from_delayed(lazy_stack,
                                         dtype=uint16,
                                         shape=self._shapes[(channel, 0)])
                      for lazy_stack in lazy_stacks]

            stacked_array = array.stack(arrays, axis=0)  # Stack all small Dask arrays into one

            return stacked_array

    def get_stack(self, channel, time_point, per_z_slice=True):

        file_name = self._get_stack_file_name(channel, time_point)
        shape = self._shapes[(channel, time_point)]

        if per_z_slice:

            lazy_get_slice_array_for_stack_file_and_z = delayed(self.cache.memoize(self._get_slice_array_for_stack_file_and_z), pure=True)

            # Lazily load each stack for each time point:
            lazy_stacks = [lazy_get_slice_array_for_stack_file_and_z(file_name, shape, z) for z in range(0, shape[0])]

            arrays = [array.from_delayed(lazy_stack,
                                         dtype=uint16,
                                         shape=shape[1:])
                      for lazy_stack in lazy_stacks]

            stack = array.stack(arrays, axis=0)

        else:
            stack = self._get_array_for_stack_file(file_name, shape=shape)

        return stack

    def add_channel(self, name: str, shape: Tuple[int, ...], dtype, enable_projections: bool = True, **kwargs) -> Any:
        raise NotImplementedError('Not implemented!')

    def get_projection_array(self, channel: str, axis: int, wrap_with_dask: bool = True) -> Any:
        array = self.get_array(channel=channel,
                               per_z_slice=False,
                               wrap_with_dask=True)
        projection = dask.array.max(array, axis=axis + 1)
        return projection

    def write_array(self, channel: str, array: numpy.ndarray):
        raise NotImplementedError('Not implemented!')

    def write_stack(self, channel: str, time_point: int, stack: numpy.ndarray):
        raise NotImplementedError('Not implemented!')

    def check_integrity(self, channels: Sequence[str]) -> bool:
        # TODO: actually implement!
        return True



import numpy
from arbol import asection

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.fusion.dct_fusion import fuse_dct_nd
from dexp.processing.fusion.dft_fusion import fuse_dft_nd
from dexp.processing.fusion.tg_fusion import fuse_tg_nd
from dexp.processing.synthetic_datasets.multiview_data import generate_fusion_test_data


def demo_fusion_numpy():
    with NumpyBackend():
        demo_fusion()


def demo_fusion_cupy():
    try:
        with CupyBackend():
            demo_fusion(include_dct=False, length_xy=512)
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")


def demo_fusion(include_dct=True, length_xy=120):
    with asection("generate data"):
        image_gt, image_lowq, blend_a, blend_b, image1, image2 = generate_fusion_test_data(add_noise=True, length_xy=length_xy, length_z_factor=4)
        image_gt = Backend.to_numpy(image_gt)

    with asection("dct fusion"):
        image_fused_dct = fuse_dct_nd(image1, image2) if include_dct else numpy.zeros_like(image_gt)
        image_fused_dct = Backend.to_numpy(image_fused_dct)

    error_dct = numpy.median(numpy.abs(image_gt - image_fused_dct))
    print(f"error_dct={error_dct}")

    with asection("dft fusion"):
        image_fused_dft = fuse_dft_nd(image1, image2)
        image_fused_dft = Backend.to_numpy(image_fused_dft)
    error_dft = numpy.median(numpy.abs(image_gt - image_fused_dft))
    print(f"error_dft={error_dft}")

    with asection("tg fusion"):
        image_fused_tg = fuse_tg_nd(image1, image2)
        image_fused_tg = Backend.to_numpy(image_fused_tg)
    error_tg = numpy.median(numpy.abs(image_gt - image_fused_tg))
    print(f"error_tg={error_tg}")

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image_gt), name='image_gt')
        viewer.add_image(_c(image_lowq), name='image_lowq')
        viewer.add_image(_c(blend_a), name='blend_a')
        viewer.add_image(_c(blend_b), name='blend_b')
        viewer.add_image(_c(image1), name='image1')
        viewer.add_image(_c(image2), name='image2')
        viewer.add_image(_c(image_fused_dct), name='image_fused_dct')
        viewer.add_image(_c(image_fused_dft), name='image_fused_dft')
        viewer.add_image(_c(image_fused_tg), name='image_fused_tg')


if __name__ == "__main__":
    # demo_fusion_cupy()
    demo_fusion_numpy()



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.interpolation.warp import warp
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data
from dexp.utils.timeit import timeit


# def test_warp_2d_numpy():
#     backend = NumpyBackend()
#     _test_warp_2d(backend)


def test_warp_3d_cupy():
    try:
        with CupyBackend():
            _test_warp_3d()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_warp_3d(length_xy=256, grid_size=8):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    with timeit("generate data"):
        _, _, image = generate_nuclei_background_data(add_noise=True,
                                                      length_xy=length_xy,
                                                      length_z_factor=1,
                                                      zoom=2,
                                                      dtype=numpy.float32)

    newimage = image[0:512, 0:511, 0:509]
    image = newimage
    image = Backend.to_backend(image)

    print(f"shape={image.shape}")

    vector_field = numpy.random.uniform(low=-5, high=+5, size=(grid_size,) * 3 + (3,))

    with timeit("warp"):
        warped = warp(image, vector_field, vector_field_upsampling=4)

    with timeit("dewarp"):
        dewarped = warp(warped, -vector_field, vector_field_upsampling=4)

    error = xp.mean(xp.absolute(image - dewarped))
    print(f"error = {error}")

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     def _c(array):
    #         return backend.to_numpy(array)
    #
    #     viewer = Viewer()
    #     viewer.add_image(_c(image), name='image', colormap='bop orange', blending='additive', rendering='attenuated_mip')
    #     viewer.add_image(_c(warped), name='warped', colormap='bop purple', blending='additive', rendering='attenuated_mip')
    #     viewer.add_image(_c(dewarped), name='dewarped', colormap='bop blue', blending='additive', rendering='attenuated_mip')
    #     viewer.camera.ndisplay = 3

    assert error < 40

    # Let's check that ndimage.shift and our warp agree:

    shifted_ndimage = sp.ndimage.shift(image, shift=(11, 5, -17))
    vector_field = xp.asarray([11, 5, -17])[numpy.newaxis, numpy.newaxis, numpy.newaxis]
    shifted_warp = warp(image, vector_field)

    error_ndimage_warp = xp.mean(xp.absolute(shifted_ndimage - shifted_warp))
    print(f"error_ndimage_warp = {error_ndimage_warp}")
    assert error_ndimage_warp < 1e-3

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     def _c(array):
    #         return backend.to_numpy(array)
    #
    #     viewer = Viewer()
    #     viewer.add_image(_c(image), name='image', colormap='bop orange', blending='additive', rendering='attenuated_mip', visible=False)
    #     viewer.add_image(_c(shifted_ndimage), name='shifted_ndimage', colormap='bop purple', blending='additive', rendering='attenuated_mip')
    #     viewer.add_image(_c(shifted_warp), name='shifted_warp', colormap='bop blue', blending='additive', rendering='attenuated_mip')
    #     viewer.camera.ndisplay = 3



import os
from os.path import exists

import click
from arbol.arbol import aprint, asection


@click.command()
@click.argument('input_path')
@click.option('--prefix', '-p', type=str, default='frame_', help='Prefix for image filenames')
@click.option('--leading', '-l', type=int, default=5, help='Number of leading zeros for frame indices.')
@click.option('--extension', '-e', type=str, default='png', help='Extension for image filenames')
@click.option('--output_path', '-o', type=str, default=None, help='Output file path for movie')
@click.option('--framerate', '-fps', type=int, default=30, help='Sets the framerate in frames per second', show_default=True)
@click.option('--codec', '-c', type=str, default='h264_nvenc',
              help='Encoding codec, For x264 class codecs: libx264 (reference but slow), or: h264_nvenc (faster, GPU). For x265 class codecs: libx265 (reference but slow), or: hevc_nvenc(faster, GPU',
              show_default=True)
@click.option('--preset', '-ps', type=str, default='slow',
              help='Possible values: ultrafast, superfast, veryfast, faster, fast, medium (default preset), slow, slower, veryslow.'
                   ' A preset is a collection of options that will provide a certain encoding speed to compression ratio.'
                   ' A slower preset will provide better compression (compression is quality per filesize).'
                   ' This means that, for example, if you target a certain file size or constant bit rate,'
                   ' you will achieve better quality with a slower preset.'
                   ' Similarly, for constant quality encoding, you will simply save bitrate by choosing a slower preset.',
              show_default=True)
@click.option('--quality', '-q', type=int, default=0,
              help='Sets the quality of the video by modulating the bit rate. Set to a higher value for higher quality, or lower (negative ok) value for lower quality. Means different things for different codecs.',
              show_default=True)
@click.option('--width', '-wi', type=int, default=0,
              help='Video frame width, recommended values: 1280(HD720), 1920(HD1080), 2048(2K), 3840(UHD-1), 4096(4K), 7680(8K). If zero then original size preserved, '
                   'if -1 the best fit is found while allowing some downscaling. Height is automatically determined to preserve aspect ratio,'
                   ' and is forced to be a multiple of 32.',
              show_default=True)
@click.option('--pixelformat', '-pf', type=str, default='yuv420p',
              help='Pixel format, can be yuv420p (default, recommended) or yuv444p (no chrominance downsampling but better compatibility),'
                   ' or any other allowed ffmpeg & codec pixel format.',
              show_default=True)
@click.option('--overwrite', '-w', is_flag=True, help='to force overwrite of target', show_default=True)  # , help='dataset slice'
def mp4(input_path,
        prefix,
        leading,
        extension,
        output_path,
        framerate,
        codec,
        preset,
        quality,
        width,
        pixelformat,
        overwrite):
    """ Converts a folder of images into an MP4 video file.
    """

    if output_path is None:
        videofilepath = os.path.basename(os.path.normpath(input_path)) + '.mp4'
    else:
        videofilepath = output_path

    videofilepath = videofilepath.replace('frames_', '')

    if overwrite or not exists(videofilepath):
        with asection(f"Converting PNG files at: {input_path}, into MP4 file: {videofilepath}, framerate: {framerate} "):

            scale_option = f'-vf scale={width}:-8:flags=bicubic' if width > 0 else ''
            # black_background_filter = f'-filter_complex "{scale_option};color=black,format={pixelformat}[c];[c][0]scale2ref[c][i];[c][i]overlay=format=auto:shortest=1,setsar=1"'

            # some codec wizardry to be able to effectively modulate quality
            quality = f'-rc vbr -cq {26 - quality} -qmin {26 - quality} -qmax {26 - quality} -b:v 0 ' if 'nvenc' in codec else f'-crf {21 - quality}'

            ffmpeg_command = f"ffmpeg -hwaccel auto -framerate {framerate} -start_number 0 -i '{input_path}/{prefix}%0{leading}d.{extension}'  " \
                             f"-f mp4 -vcodec {codec} -preset {preset}  {quality} -pix_fmt {pixelformat} {scale_option} -y {videofilepath}"
            # -vf  \"crop=576:1240:320:0\"  ""  ,setsar=1:1
            # -pix_fmt {pixelformat} {scale_option}
            # ,setsar=1:1
            # h264_nvenc nvenc_h264 libx264

            aprint(f"Executing command: '{ffmpeg_command}'")

            os.system(ffmpeg_command)
    else:
        aprint(f"Video file: {videofilepath} already exists! use -w option to force overwrite...")
#
# #low quality
# #ffmpeg -r 4 -i /data_fish_TL100_range1200um_step0.31_4um_15ms_dualv_fused_colormax_projection/frame_%05d.png -vcodec mpeg4 -y movie.mp4
#
# #high quality
# ffmpeg -framerate 30 -i data_fish_TL100_range1200um_step0.31_4um_15ms_dualv_fused_stabilized_colormax_projection/frame_%05d.png -f mp4 -vcodec libx264 -preset slow -pix_fmt yuv420p -y movie.mp4
#
# #change resolution
# ffmpeg -i movie.mp4 -s hd1080 movie_1080.mp4
# # or use -vf scale=1920x1080 to set to other scales



import numpy
from skimage.data import camera
from skimage.util import random_noise

from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.filters.butterworth_filter import butterworth_filter


def test_butterworth_numpy():
    with NumpyBackend():
        _test_butterworth()


def test_butterworth_cupy():
    try:
        with CupyBackend():
            _test_butterworth()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_butterworth():
    image = camera().astype(numpy.float32) / 255
    noisy = random_noise(image.copy(), mode="gaussian", var=0.005, seed=0, clip=False)
    # noisy = random_noise(noisy, mode="s&p", amount=0.03, seed=0, clip=False)

    filtered = butterworth_filter(noisy)

    assert filtered.shape == noisy.shape
    assert filtered.dtype == noisy.dtype

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     viewer = Viewer()
    #     viewer.add_image(image, name='image')
    #     viewer.add_image(noisy, name='noisy')
    #     viewer.add_image(filtered, name='filtered')



from typing import Sequence, Optional

import numpy
from arbol.arbol import aprint
from skimage.transform import downscale_local_mean

from dexp.datasets.base_dataset import BaseDataset
from dexp.utils.timeit import timeit


def dataset_isonet(dataset: BaseDataset,
                   path: str,
                   channel: Optional[Sequence[str]],
                   slicing,
                   store: str,
                   compression: str,
                   compression_level: int,
                   overwrite: bool,
                   context: str,
                   mode: str,
                   dxy: float,
                   dz: float,
                   binning: int,
                   sharpening: bool,
                   training_tp_index: Optional[int],
                   max_epochs: int,
                   check: bool):
    if channel is None:
        channel = 'fused'

    if training_tp_index is None:
        training_tp_index = dataset.nb_timepoints(channel) // 2

    aprint(f"Selected channel {channel}")

    aprint(f"getting Dask arrays to apply isonet on...")
    array = dataset.get_array(channel, per_z_slice=False, wrap_with_dask=True)

    if slicing is not None:
        aprint(f"Slicing with: {slicing}")
        array = array[slicing]

    aprint(f"Binning image by a factor {binning}...")
    dxy *= binning

    subsampling = dz / dxy
    aprint(f"Parameters: dxy={dxy}, dz={dz}, subsampling={subsampling}")

    psf = numpy.ones((1, 1)) / 1
    print(f"PSF (along xy): {psf}")
    from dexp.processing.isonet import IsoNet
    isonet = IsoNet(context, subsampling=subsampling)

    if 'p' in mode:
        training_array_tp = array[training_tp_index].compute()
        training_downscaled_array_tp = downscale_local_mean(training_array_tp, factors=(1, binning, binning))
        print(f"Training image shape: {training_downscaled_array_tp.shape} ")
        isonet.prepare(training_downscaled_array_tp, psf=psf, threshold=0.999)

    if 't' in mode:
        isonet.train(max_epochs=max_epochs)

    if 'a' in mode:
        from dexp.datasets.zarr_dataset import ZDataset
        mode = 'w' + ('' if overwrite else '-')
        dest_dataset = ZDataset(path, mode, store)
        zarr_array = None

        for tp in range(0, array.shape[0] - 1):
            with timeit('Elapsed time: '):

                aprint(f"Processing time point: {tp} ...")
                tp_array = array[tp].compute()

                aprint("Downscaling image...")
                tp_array = downscale_local_mean(tp_array, factors=(1, binning, binning))
                # array_tp_downscaled = zoom(tp_array, zoom=(1, 1.0/binning, 1.0/binning), order=0)

                if sharpening:
                    aprint("Sharpening image...")
                    from dexp.processing.restoration import dehazing
                    tp_array = dehazing(tp_array, mode='hybrid', min=0, max=1024, margin_pad=False)

                aprint("Applying IsoNet to image...")
                tp_array = isonet.apply(tp_array)

                aprint(f'Result: image of shape: {tp_array.shape}, dtype: {tp_array.dtype} ')

                if zarr_array is None:
                    shape = (array.shape[0],) + tp_array.shape
                    aprint(f"Creating Zarr array of shape: {shape} ")
                    zarr_array = dest_dataset.add_channel(name=channel,
                                                          shape=shape,
                                                          dtype=array.dtype,
                                                          chunks=ZDataset._default_chunks,
                                                          codec=compression,
                                                          clevel=compression_level)

                aprint(f'Writing image to Zarr file...')
                dest_dataset.write_stack(channel=channel,
                                         time_point=tp,
                                         stack_array=tp_array)

                aprint(f"Done processing time point: {tp}")

        aprint(dest_dataset.info())
        if check:
            dest_dataset.check_integrity()

        # close destination dataset:
        dest_dataset.close()



import random

from arbol import aprint, asection
from skimage.data import camera

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.interpolation.warp import warp
from dexp.processing.registration.sequence import image_stabilisation
from dexp.processing.synthetic_datasets.binary_blobs import binary_blobs
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data


def demo_register_sequence_2d_numpy():
    with NumpyBackend():
        _register_sequence_2d()


def demo_register_sequence_2d_cupy():
    try:
        with CupyBackend():
            _register_sequence_2d()
    except ModuleNotFoundError:
        aprint("Cupy module not found! demo ignored")


def _register_sequence_2d(length_xy=512,
                          n=512,
                          drift_strength=1.5,
                          warp_grid_size=6,
                          warp_strength=2.5,
                          ratio_bad_frames=0.05,
                          additive_noise=0.05,
                          multiplicative_noise=0.1,
                          run_test=False,
                          display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    # used for simulation:
    shifts = []
    vector_fields = []
    x, y = 0, 0
    vector_field = xp.zeros((warp_grid_size,) * 2 + (2,), dtype=xp.float32)

    with asection("generate dataset"):

        random.seed(0)
        xp.random.seed(0)

        # prepare simulation data:
        for i in range(n):
            # drift:
            if i < (2 * n // 3):
                x += drift_strength * random.uniform(-1, 1.7)
                y += drift_strength * random.uniform(-1.3, 1)
            else:
                x += drift_strength * random.uniform(-1.5, 1)
                y += drift_strength * random.uniform(-1, 1.5)

            # deformation over time:
            vector_field += warp_strength * xp.random.uniform(low=-1, high=+1, size=(warp_grid_size,) * 2 + (2,))

            # simulate sudden imaging jumps:
            if i == n // 2 - n // 3:
                x += 38
                y += -35
                vector_field += 16 * warp_strength * xp.random.uniform(low=-1, high=+1, size=(warp_grid_size,) * 2 + (2,))
            if i == n // 2 + n // 3:
                x += 37
                y += -29
                vector_field += 16 * warp_strength * xp.random.uniform(low=-1, high=+1, size=(warp_grid_size,) * 2 + (2,))

            # keep for later:
            shifts.append((x, y))
            vector_fields.append(xp.copy(vector_field))

        # blobs:
        image1 = binary_blobs(length=length_xy, seed=1, n_dim=2, blob_size_fraction=0.04, volume_fraction=0.05)
        image1 = image1.astype(xp.float32)
        image1 = sp.ndimage.gaussian_filter(image1, sigma=4)

        # camera:
        image2 = camera().astype(xp.float32)

        # nuclei:
        _, _, image3 = generate_nuclei_background_data(add_noise=False,
                                                       length_xy=length_xy // 4,
                                                       length_z_factor=1,
                                                       independent_haze=True,
                                                       sphere=True,
                                                       add_offset=False,
                                                       zoom=4,
                                                       dtype=xp.float32)
        image3 = image3[image3.shape[0] // 2].astype(xp.float32)

        # choice between 3 images:
        image = Backend.to_backend(image3)

        # crop so we test weird non-square image shape:
        image = image[0:511, 0:501]

        # generate reference 'ground truth' timelapse
        image = xp.stack(image for _ in range(n))

        # generate shifted, distorted, degraded timelapse:
        shifted = xp.stack((sp.ndimage.shift(warp(i, vf, vector_field_upsampling=4), shift=s) for i, s, vf in zip(image, shifts, vector_fields)))
        shifted *= xp.clip(xp.random.normal(loc=1, scale=multiplicative_noise, size=shifted.shape, dtype=xp.float32), 0.1, 10)
        shifted += additive_noise * xp.random.rand(*shifted.shape, dtype=xp.float32)
        shifted = xp.clip(shifted - 50, 0)

        # simulate dropped, highly corrupted frames:
        for _ in range(int(shifted.shape[0] * ratio_bad_frames)):
            index = random.randint(0, n - 1)
            shifted[index] = sp.ndimage.shift(shifted[index], shift=(random.uniform(-27, 27), random.uniform(-27, 27)))
            shifted[index] += xp.random.rand(*shifted[index].shape)
            shifted[index] *= random.uniform(0.001, 0.5)
            shifted[index] += random.uniform(0.001, 0.1) * xp.random.rand(*shifted[index].shape)

    with asection("register_translation_2d"):
        # compute image sequence stabilisation model:
        model = image_stabilisation(shifted,
                                    axis=0,
                                    sigma=3,
                                    edge_filter=False,
                                    denoise_input_sigma=1,
                                    debug_output='stabilisation')
        aprint(f"model: {model}")

    with asection("shift back"):

        # apply stabilisation:
        stabilised_seq = model.apply_sequence(shifted, axis=0)

        # another way to apply stabilisation:
        stabilised_sps = xp.stack(model.apply(image, index=i, pad=True) for i, image in enumerate(shifted))

    if display:
        from napari import Viewer, gui_qt
        with gui_qt():
            def _c(array):
                return Backend.to_numpy(array)

            viewer = Viewer()
            viewer.grid.enabled = True
            # viewer.add_image(_c(image), name='image', colormap='bop orange', blending='additive', visible=True)
            viewer.add_image(_c(shifted), name='shifted', colormap='bop purple', blending='additive', visible=True)
            viewer.add_image(_c(stabilised_seq), name='stabilised_seq', colormap='bop blue', blending='additive', visible=True)
            # viewer.add_image(_c(stabilised_sps), name='stabilised_sps', colormap='bop blue', blending='additive', visible=True)

    return image, shifted, stabilised_seq, model


if __name__ == "__main__":
    demo_register_sequence_2d_cupy()
    # demo_register_translation_2d_numpy()



import gc
import math
import os
import threading
from typing import Any

import numpy
from arbol import aprint
from dask.array import Array

from dexp.processing.backends.backend import Backend


class CupyBackend(Backend):
    """
    CupyBackend
    """
    _dexp_cuda_cluster = None
    _dexp_dask_client = None

    @staticmethod
    def num_devices():
        import GPUtil
        return len(GPUtil.getGPUs())

    @staticmethod
    def available_devices(order='first', maxLoad=math.inf, maxMemory=math.inf):
        """

        Parameters
        ----------
        order :     first --> select the GPU with the lowest ID (DEFAULT)
                    last --> select the GPU with the highest ID
                    random --> select a random available GPU
                    load --> select the GPU with the lowest load
                    memory --> select the GPU with the most memory available
        maxLoad :
        maxMemory

        Returns
        -------

        """
        # Set CUDA_DEVICE_ORDER so the IDs assigned by CUDA match those from nvidia-smi
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        import GPUtil
        return GPUtil.getAvailable(order=order,
                                   limit=math.inf,
                                   maxLoad=maxLoad,
                                   maxMemory=maxMemory,
                                   includeNan=False,
                                   excludeID=[],
                                   excludeUUID=[])

    device_locks = tuple(threading.Lock() for _ in range(num_devices.__func__()))

    def __init__(self,
                 device_id=0,
                 exclusive: bool = False,
                 enable_streaming: bool = True,
                 enable_memory_pool: bool = True,
                 enable_memory_pool_clearing: bool = True,
                 enable_unified_memory: bool = True,
                 enable_cub: bool = True,
                 enable_cutensor: bool = True,
                 enable_fft_planning: bool = True,
                 enable_dask_cuda_cluster: bool = False,
                 enable_dask_cuda_nvlink: bool = False,
                 ):
        """
        Instantiates a Cupy-based Image Processing backend

        Parameters
        ----------
        device_id : CUDA device id to use for allocation and compute
        exclusive : If True the access to this device is exclusive, no other backend context can access it (when using the context manager idiom)
        enable_memory_pool : Enables cupy memory pool. By default disabled, when enabled cupy tends to return out-of-memory exceptions when handling large arrays.
        enable_memory_pool_clearing : Enables the clearing of the memory pool upon calling 'clear_memory_pool'
        enable_cub : enables CUB accelerator
        enable_cutensor : enables cuTensor accelerator
        enable_fft_planning : enables FFT planning
        enable_dask_cuda_cluster : enables dask cuda cluster
        enable_dask_cuda_nvlink : enables nvlink
        """

        super().__init__()
        self.device_id = device_id
        self.exclusive = exclusive
        self.enable_memory_pool = enable_memory_pool
        self.enable_memory_pool_clearing = enable_memory_pool_clearing
        self.enable_streaming = enable_streaming
        self.enable_unified_memory = enable_unified_memory
        self.enable_cub = enable_cub
        self.enable_cutensor = enable_cutensor
        self.enable_fft_planning = enable_fft_planning
        self.enable_dask_cuda_cluster = enable_dask_cuda_cluster
        self.enable_dask_cuda_nvlink = enable_dask_cuda_nvlink

        self.mempool = None
        self.stream = None
        self._previous_allocator = None

        import cupy
        self.cupy_device: cupy.cuda.Device = cupy.cuda.Device(self.device_id)

        from cupy.cuda import cub, cutensor
        cub.available = enable_cub
        cutensor.available = enable_cutensor

        try:
            import cupy.cudnn
            aprint("CUDNN available!")
        except Exception as e:
            pass

        if not enable_fft_planning:
            import cupy
            cupy.fft.config.enable_nd_planning = False

        if enable_dask_cuda_cluster and CupyBackend._dexp_cuda_cluster is None:
            from dask_cuda import LocalCUDACluster
            from distributed import Client

            CupyBackend._dexp_cuda_cluster = LocalCUDACluster(enable_nvlink=enable_dask_cuda_nvlink)
            CupyBackend._dexp_dask_client = Client(CupyBackend._dexp_cuda_cluster)

        ## Important: Leave this, this is to make sure that some package works properly!
        exec("import cupyx.scipy.ndimage")
        exec("import cupyx.scipy.special")
        exec("import scipy.special")

    def copy(self, exclusive: bool = None):
        return CupyBackend(device_id=self.device_id,
                           exclusive=self.exclusive if exclusive is None else exclusive,
                           enable_streaming=self.enable_streaming,
                           enable_memory_pool=self.enable_memory_pool,
                           enable_memory_pool_clearing=self.enable_memory_pool_clearing,
                           enable_unified_memory=self.enable_unified_memory,
                           enable_cub=self.enable_cub,
                           enable_cutensor=self.enable_cutensor,
                           enable_fft_planning=self.enable_fft_planning,
                           enable_dask_cuda_cluster=self.enable_dask_cuda_cluster,
                           enable_dask_cuda_nvlink=self.enable_dask_cuda_nvlink)

    def __str__(self):
        free_mem = self.cupy_device.mem_info[0]
        total_mem = self.cupy_device.mem_info[0]
        percent = (100 * free_mem) // total_mem
        return (f"Cupy backend [device id:{self.device_id} "
                f"with {free_mem // (1024 * 1024)} MB ({percent}%) free memory out of {free_mem // (1024 * 1024)} MB, "
                f"compute:{self.cupy_device.compute_capability}, pci-bus-id:'{self.cupy_device.pci_bus_id}']")

    def __enter__(self):

        # lock device:
        if self.exclusive:
            CupyBackend.device_locks[self.device_id].acquire(blocking=True)

        # setup device:
        self.cupy_device.__enter__()

        # setup allocation:
        if self.enable_memory_pool:
            if self.mempool is None:
                import cupy
                from cupy.cuda.memory import SingleDeviceMemoryPool
                self.mempool = SingleDeviceMemoryPool(cupy.cuda.memory.malloc_managed if self.enable_unified_memory else None)
                from cupy.cuda import memory
            self._previous_allocator = memory._get_thread_local_allocator()
            memory._set_thread_local_allocator(self.mempool.malloc)
        else:
            from cupy.cuda import memory
            memory._set_thread_local_allocator(None)

        # setup stream:
        if self.enable_streaming:
            import cupy
            self.stream = cupy.cuda.stream.Stream(non_blocking=True)
            self.stream.__enter__()

        return super().__enter__()

    def __exit__(self, type, value, traceback):

        super().__exit__(type, value, traceback)

        # unset stream:
        if self.enable_streaming and self.stream is not None:
            self.stream.synchronize()
            self.stream.__exit__()

        # unset allocation:
        self.clear_memory_pool()
        if self._previous_allocator is not None:
            from cupy.cuda import memory
            memory._set_thread_local_allocator(self._previous_allocator)

        # unset device:
        self.cupy_device.__exit__()

        # unlock device:
        if self.exclusive:
            CupyBackend.device_locks[self.device_id].release()

    def synchronise(self):
        if self.stream is not None:
            self.stream.synchronize()

    def clear_memory_pool(self):

        if self.enable_memory_pool_clearing:
            if self.mempool is not None:
                n_free_blocks_before = self.mempool.n_free_blocks()
                used_before = self.mempool.used_bytes() // 1e9
                gc.collect()
                self.mempool.free_all_blocks(self.stream)
                n_free_blocks_after = self.mempool.n_free_blocks()
                used_after = self.mempool.used_bytes() // 1e9
                total_after = self.mempool.total_bytes() // 1e9

                aprint(f"Number of free blocks before and after release: {n_free_blocks_before}->{n_free_blocks_after}, used:{used_before}GB->{used_after}GB, total:{total_after}GB ")
            else:
                aprint("Warning: default cupy memory pool is 'None'")
        else:
            aprint("Memory pool clearing not enabled!")

        super().clear_memory_pool()

    def _to_numpy(self, array, dtype=None, force_copy: bool = False) -> numpy.ndarray:
        import cupy

        if isinstance(array, Array):
            return self._to_numpy(array.compute(), dtype=dtype, force_copy=force_copy)

        elif cupy.get_array_module(array) == cupy:
            array = cupy.asnumpy(array)

        if dtype:
            return numpy.asarray(array).astype(dtype, copy=force_copy)
        elif force_copy:
            return numpy.asarray(array.copy())
        else:
            return numpy.asarray(array)

    def _to_backend(self, array, dtype=None, force_copy: bool = False) -> Any:

        import cupy

        if isinstance(array, Array):
            return self._to_backend(array.compute(), dtype=dtype, force_copy=force_copy)

        elif cupy.get_array_module(array) == cupy:
            if dtype:
                return array.astype(dtype, copy=force_copy)
            elif force_copy:
                return array.copy()
            else:
                return array
        else:
            array = self.to_numpy(array)
            with self.cupy_device:
                return cupy.asarray(array, dtype=dtype)

    def _get_xp_module(self, array=None) -> Any:
        if array is None:
            import cupy
            return cupy
        else:
            import cupy
            return cupy.get_array_module(array)

    def _get_sp_module(self, array=None) -> Any:
        if array is None:
            import cupyx
            return cupyx.scipy
        else:
            import cupyx
            return cupyx.scipy.get_array_module(array)



from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.color.demo.demo_border import demo_border
from dexp.processing.color.demo.demo_crop_resize_pad import demo_crop_resize_pad


def test_crop_resize_pad_numpy():
    with NumpyBackend():
        _test_crop_resize_pad()


def test_crop_resize_pad_cupy():
    try:
        with CupyBackend():
            _test_crop_resize_pad()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_crop_resize_pad():
    demo_crop_resize_pad(display=False)



# You need to point to a tiff file with 4 views as first dim,
# as produced for example by: dexp tiff -w -s [128:129] dataset.zarr -o /home/royer/Desktop/test_data/test_data.tiff
from arbol import asection, aprint
from napari import Viewer

from dexp.datasets.zarr_dataset import ZDataset
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.multiview_lightsheet.fusion.mvsols import msols_fuse_1C2L

# dataset_path = '/mnt/raid0/pisces_datasets/data2_fish_TL100_range1300um_step0.31_6um_20ms_dualv_300tp_2_first10tp.zarr'
dataset_path = '/mnt/raid0/dexp_datasets/tail/raw.zarr'


def demo_mvsols_resample_numpy():
    with NumpyBackend():
        _mvsols_resample()


def demo_mvsols_resample_cupy():
    try:
        with CupyBackend():
            _mvsols_resample()
    except ModuleNotFoundError:
        print("Cupy module not found! demo ignored")


def _mvsols_resample():
    xp = Backend.get_xp_module()

    with asection(f"Load"):
        zdataset = ZDataset(path=dataset_path, mode='r')

        aprint(zdataset.channels())

        C0L0 = zdataset.get_stack('v0c0', 0)
        C0L1 = zdataset.get_stack('v1c0', 0)

        aprint(f"C0L0 shape={C0L0.shape}, dtype={C0L0.dtype}")
        aprint(f"C0L1 shape={C0L1.shape}, dtype={C0L1.dtype}")

        metadata = zdataset.get_metadata()
        aprint(metadata)

    with asection(f"Fuse"):
        angle = metadata['angle']
        channel = metadata['channel']
        dz = metadata['dz']
        res = metadata['res']

        C0Lx, model, ratios = msols_fuse_1C2L(C0L0, C0L1,
                                              zero_level=0,
                                              angle=angle,
                                              dx=res,
                                              dz=dz,
                                              equalise=True,
                                              registration_confidence_threshold=0.6,
                                              registration_max_residual_shift=64,
                                              z_pad=8,
                                              z_apodise=96)
        aprint(f"model = {model}, ratios={ratios}")

    from napari import gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        # viewer.add_image(_c(C0L0), name='C0L0', colormap='bop blue', blending='additive')
        # viewer.add_image(_c(C0L1), name='C0L1', colormap='bop orange', blending='additive')
        viewer.add_image(_c(C0Lx), name='C0Lx', colormap='bop blue', blending='additive')


if __name__ == "__main__":
    # demo_simview_fuse_numpy()
    demo_mvsols_resample_cupy()



import numpy
from arbol import aprint, asection
from skimage.data import camera

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.interpolation.warp import warp
from dexp.processing.registration.warp_multiscale_nd import register_warp_multiscale_nd


def demo_register_warp_2d_ms_numpy():
    with NumpyBackend():
        _register_warp_2d_ms()


def demo_register_warp_2d_ms_cupy():
    try:
        with CupyBackend():
            _register_warp_2d_ms()
    except ModuleNotFoundError:
        aprint("Cupy module not found! demo ignored")


def _register_warp_2d_ms(warp_grid_size=4, reg_grid_size=8, display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    with asection("generate dataset"):
        image = camera().astype(numpy.float32)
        image = image[0:510, 0:509]
        image = Backend.to_backend(image)

    with asection("warp"):
        magnitude = 20
        vector_field = numpy.random.uniform(low=-magnitude, high=+magnitude, size=(warp_grid_size,) * 2 + (2,))
        warped = warp(image, vector_field, vector_field_upsampling=4)
        aprint(f"vector field applied: {vector_field}")

    with asection("add noise"):
        image += xp.random.uniform(0, 20, size=image.shape)
        warped += xp.random.uniform(0, 20, size=warped.shape)

    with asection("register_warp_multiscale_nd"):
        model = register_warp_multiscale_nd(image, warped,
                                            num_iterations=5,
                                            confidence_threshold=0.3,
                                            edge_filter=False)

        aprint(f"vector field found: {model.vector_field}")

    with asection("unwarp"):
        _, unwarped = model.apply_pair(image, warped, vector_field_upsampling=4)

    if display:
        from napari import Viewer, gui_qt
        with gui_qt():
            def _c(array):
                return Backend.to_numpy(array)

            viewer = Viewer()
            viewer.add_image(_c(image), name='image', colormap='bop orange', blending='additive')
            viewer.add_image(_c(warped), name='warped', colormap='bop blue', blending='additive', visible=False)
            viewer.add_image(_c(unwarped), name='unwarped', colormap='bop purple', blending='additive')

    return image, warped, unwarped, model


if __name__ == "__main__":
    demo_register_warp_2d_ms_cupy()
    # demo_register_warp_2D_numpy()



import os
from typing import Sequence

import numpy
from arbol.arbol import aprint
from arbol.arbol import asection
from joblib import Parallel, delayed

from dexp.datasets.base_dataset import BaseDataset
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.best_backend import BestBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.registration.sequence_proj import image_stabilisation_proj_


def dataset_stabilize(input_dataset: BaseDataset,
                      output_path: str,
                      channels: Sequence[str],
                      slicing=None,
                      zarr_store: str = 'dir',
                      compression_codec: str = 'zstd',
                      compression_level: int = 3,
                      overwrite: bool = False,
                      max_range: int = 7,
                      min_confidence: float = 0.5,
                      enable_com: bool = False,
                      quantile: float = 0.5,
                      tolerance: float = 1e-7,
                      order_error: float = 2.0,
                      order_reg: float = 1.0,
                      alpha_reg: float = 0.1,
                      phase_correlogram_sigma: float = 2,
                      denoise_input_sigma: float = 1.5,
                      log_compression: bool = True,
                      edge_filter: bool = False,
                      pad: bool = True,
                      integral: bool = True,
                      workers: int = -1,
                      workers_backend: str = 'threading',
                      device: int = 0,
                      check: bool = True,
                      stop_at_exception: bool = True,
                      debug_output=None):
    """

    Takes an input dataset and performs image stabilisation and outputs a stabilised dataset with given selected slice & channels in Zarr format with given store type, compression, etc...

    Parameters
    ----------
    input_dataset: Input dataset
    output_path: Output path for Zarr storage
    channels: selected channels
    slicing: selected array slicing
    zarr_store: type of store, can be 'dir', 'ndir', or 'zip'
    compression_codec: compression codec to be used ('zstd', 'blosclz', 'lz4', 'lz4hc', 'zlib' or 'snappy').
    compression_level: An integer between 0 and 9 specifying the compression level.
    overwrite: overwrite output dataset if already exists
    max_range: maximal distance, in time points, between pairs of images to registrate.
    min_confidence: minimal confidence below which pairwise registrations are rejected for the stabilisation.
    enable_com: enable center of mass fallback when standard registration fails.
    quantile: quantile to cut-off background in center-of-mass calculation
    tolerance: tolerance for linear solver.
    order_error: order for linear solver error term.
    order_reg: order for linear solver regularisation term.
    alpha_reg: multiplicative coefficient for regularisation term.
    phase_correlogram_sigma: sigma for Gaussian smoothing of phase correlogram, zero to disable.
    denoise_input_sigma : Uses a Gaussian filter to denoise input images, zero to disable.
    log_compression : Applies the function log1p to the images to compress high-intensities (usefull when very (too) bright structures are present in the images, such as beads)
    edge_filter : apply sobel edge filter to input images.
    pad: pad input dataset.
    integral: Set to True to only allow integral translations, False to allow subpixel accurate translations (induces blur!).
    workers: number of workers, if -1 then the number of workers == number of cores or devices
    workers_backend: What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread)
    device: Sets the CUDA devices id, e.g. 0,1,2
    check: Checking integrity of written file.
    stop_at_exception: True to stop as soon as there is an exception during processing.
    """

    from dexp.datasets.zarr_dataset import ZDataset
    mode = 'w' + ('' if overwrite else '-')
    output_dataset = ZDataset(output_path, mode, zarr_store)

    for channel in input_dataset._selected_channels(channels):

        # get channel array:
        array = input_dataset.get_array(channel, per_z_slice=False, wrap_with_dask=True)

        # Perform slicing:
        if slicing is not None:
            array = array[slicing]

        # Shape and chunks for array:
        ndim = array.ndim
        shape = array.shape
        dtype = array.dtype
        chunks = ZDataset._default_chunks
        nb_timepoints = shape[0]

        # Obtain projections:
        projections = []
        for axis in range(array.ndim - 1):
            projection = input_dataset.get_projection_array(channel=channel, axis=axis, wrap_with_dask=False)
            if slicing is not None:
                projection = projection[slicing]

            proj_axis = list(1 + a for a in range(array.ndim - 1) if a != axis)
            projections.append((*proj_axis, projection))

        # Perform stabilisation:
        with BestBackend(device, enable_unified_memory=True):
            model = image_stabilisation_proj_(projections=projections,
                                              max_range=max_range,
                                              min_confidence=min_confidence,
                                              enable_com=enable_com,
                                              quantile=quantile,
                                              tolerance=tolerance,
                                              order_error=order_error,
                                              order_reg=order_reg,
                                              alpha_reg=alpha_reg,
                                              sigma=phase_correlogram_sigma,
                                              denoise_input_sigma=denoise_input_sigma,
                                              log_compression=log_compression,
                                              edge_filter=edge_filter,
                                              ndim=ndim - 1,
                                              internal_dtype=numpy.float16,
                                              debug_output=debug_output)
            model.to_numpy()

        # Shape of the resulting array:
        padded_shape = (nb_timepoints,) + model.padded_shape(shape[1:])

        # Add channel to output datatset:
        output_dataset.add_channel(name=channel,
                                   shape=padded_shape,
                                   dtype=dtype,
                                   chunks=chunks,
                                   codec=compression_codec,
                                   clevel=compression_level)

        # definition of function that processes each time point:
        def process(tp):

            try:
                with asection(f"Processing time point: {tp}/{nb_timepoints} ."):
                    with asection(f"Loading stack"):
                        tp_array = array[tp].compute()

                    # backend = NumpyBackend() if integral else CupyBackend(device, exclusive=True)

                    with NumpyBackend():
                        xp = Backend.get_xp_module()

                        with asection(f"Applying model..."):
                            # tp_array = Backend.to_backend(tp_array, dtype=array.dtype, force_copy=False)
                            tp_array = model.apply(tp_array, index=tp, pad=pad, integral=integral)
                            # tp_array = Backend.to_numpy(tp_array, dtype=array.dtype, force_copy=False)

                    with asection(f"Saving stabilized stack for time point {tp}/{nb_timepoints}, shape:{tp_array.shape}, dtype:{array.dtype}"):
                        output_dataset.write_stack(channel=channel,
                                                   time_point=tp,
                                                   stack_array=tp_array)



            except Exception as error:
                aprint(error)
                aprint(f"Error occurred while processing time point {tp}/{nb_timepoints} !")
                import traceback
                traceback.print_exc()
                if stop_at_exception:
                    raise error

        # Set number of workers:
        if workers == -1:
            workers = max(1, os.cpu_count() // abs(workers))
        aprint(f"Number of workers: {workers}")

        # start jobs:
        if workers > 1:
            Parallel(n_jobs=workers, backend=workers_backend)(delayed(process)(tp) for tp in range(0, nb_timepoints))
        else:
            for tp in range(0, nb_timepoints):
                process(tp)

    # printout output dataset info:
    aprint(output_dataset.info())
    if check:
        output_dataset.check_integrity()

    # close destination dataset:
    output_dataset.close()



from dexp.processing.backends.backend import Backend


def gray2rgba(image,
              alpha=None,
              rgba_value_max: float = 255):
    """Create a RGBA representation of a gray-level image.

    Parameters
    ----------
    image : array_like
        Input image.
    alpha : array_like, optional
        Alpha channel of the output image. It may be a scalar or an
        array that can be broadcast to ``image``. If not specified it is
        set to the maximum limit corresponding to the ``image`` dtype.

    Returns
    -------
    rgba : ndarray
        RGBA image. A new dimension of length 4 is added to input
        image shape.
    """

    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    arr = xp.asarray(image)

    if alpha is None:
        alpha = rgba_value_max

    rgba = xp.empty(arr.shape + (4,), dtype=arr.dtype)
    rgba[..., :3] = arr[..., xp.newaxis]
    rgba[..., 3] = alpha

    return rgba
